# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Objects and relations for information capture and data modelling.

Information objects inheritance tree (taxonomy):

.. diagram:: [CDB] Information [Taxonomy]

Information object-relations map (ontology):

.. diagram:: [CDB] Information [Ontology]
"""

from __future__ import annotations

import capellambse.model as m

from .. import capellacommon, capellacore, modellingcore, modeltypes
from . import datatype, datavalue


@m.xtype_handler(None)
class Unit(m.ModelElement):
    """Unit."""

    _xmltag = "ownedUnits"


@m.xtype_handler(None)
class Association(m.ModelElement):
    """An Association."""

    _xmltag = "ownedAssociations"

    members: m.Accessor[Property]
    navigable_members: m.Accessor[Property]

    @property
    def roles(self) -> m.ElementList[Property]:
        assert isinstance(self.members, m.ElementList)
        assert isinstance(self.navigable_members, m.ElementList)
        roles = [i._element for i in self.members + self.navigable_members]
        return m.ElementList(self._model, roles, Property)


@m.xtype_handler(None)
class PortAllocation(modellingcore.TraceableElement):
    """An exchange between a ComponentPort and FunctionalPort."""

    _xmltag = "ownedPortAllocations"


@m.xtype_handler(None)
class Property(m.ModelElement):
    """A Property of a Class."""

    _xmltag = "ownedFeatures"

    is_ordered = m.BoolPOD("ordered")
    """Indicates if property is ordered."""
    is_unique = m.BoolPOD("unique")
    """Indicates if property is unique."""
    is_abstract = m.BoolPOD("isAbstract")
    """Indicates if property is abstract."""
    is_static = m.BoolPOD("isStatic")
    """Indicates if property is static."""
    is_part_of_key = m.BoolPOD("isPartOfKey")
    """Indicates if property is part of key."""
    is_derived = m.BoolPOD("isDerived")
    """Indicates if property is abstract."""
    is_read_only = m.BoolPOD("isReadOnly")
    """Indicates if property is read-only."""
    visibility = m.EnumPOD(
        "visibility", modeltypes.VisibilityKind, default="UNSET"
    )
    kind = m.EnumPOD(
        "aggregationKind", modeltypes.AggregationKind, default="UNSET"
    )
    type = m.AttrProxyAccessor(m.ModelElement, "abstractType")
    default_value = m.RoleTagAccessor("ownedDefaultValue")
    min_value = m.RoleTagAccessor("ownedMinValue")
    max_value = m.RoleTagAccessor("ownedMaxValue")
    null_value = m.RoleTagAccessor("ownedNullValue")
    min_card = m.RoleTagAccessor("ownedMinCard")
    max_card = m.RoleTagAccessor("ownedMaxCard")
    association = m.ReferenceSearchingAccessor(Association, "roles")


@m.xtype_handler(None)
class Class(m.ModelElement):
    """A Class."""

    _xmltag = "ownedClasses"

    sub: m.Accessor
    super: m.Accessor[Class]
    is_abstract = m.BoolPOD("abstract")
    """Indicates if class is abstract."""
    is_final = m.BoolPOD("final")
    """Indicates if class is final."""
    is_primitive = m.BoolPOD("isPrimitive")
    """Indicates if class is primitive."""
    visibility = m.EnumPOD(
        "visibility", modeltypes.VisibilityKind, default="UNSET"
    )
    state_machines = m.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=m.ElementList
    )
    owned_properties = m.DirectProxyAccessor(Property, aslist=m.ElementList)
    generalizations = m.DirectProxyAccessor(
        capellacore.Generalization, aslist=m.ElementList
    )

    @property
    def properties(self) -> m.ElementList[Property]:
        """Return all owned and inherited properties."""
        return (
            self.owned_properties + self.super.properties
            if self.super is not None
            else self.owned_properties
        )


@m.xtype_handler(None)
class InformationRealization(modellingcore.TraceableElement):
    """A realization for a Class."""

    _xmltag = "ownedInformationRealizations"


@m.xtype_handler(None)
class Union(Class):
    """A Union."""

    _xmltag = "ownedClasses"

    kind = m.EnumPOD("kind", modeltypes.UnionKind, default="UNION")


@m.xtype_handler(None)
class Collection(m.ModelElement):
    """A Collection."""

    _xmltag = "ownedCollections"

    kind = m.EnumPOD("kind", modeltypes.CollectionKind, default="ARRAY")

    sub: m.Accessor
    super: m.Accessor[Collection]


@m.xtype_handler(None)
class DataPkg(m.ModelElement):
    """A data package that can hold classes."""

    _xmltag = "ownedDataPkgs"

    owned_associations = m.DirectProxyAccessor(
        Association, aslist=m.ElementList
    )
    classes = m.DirectProxyAccessor(Class, aslist=m.ElementList)
    unions = m.DirectProxyAccessor(Union, aslist=m.ElementList)
    collections = m.DirectProxyAccessor(Collection, aslist=m.ElementList)
    enumerations = m.DirectProxyAccessor(
        datatype.Enumeration, aslist=m.ElementList
    )
    datatypes = m.DirectProxyAccessor(
        m.ModelElement,
        (
            datatype.BooleanType,
            datatype.Enumeration,
            datatype.StringType,
            datatype.NumericType,
            datatype.PhysicalQuantity,
        ),
        aslist=m.MixedElementList,
    )
    complex_values = m.DirectProxyAccessor(
        datavalue.ComplexValue, aslist=m.ElementList
    )
    packages: m.Accessor


@m.xtype_handler(None)
class ExchangeItemElement(m.ModelElement):
    """An ExchangeItemElement (proxy link)."""

    _xmltag = "ownedElements"

    abstract_type = m.AttrProxyAccessor(m.ModelElement, "abstractType")
    owner = m.ParentAccessor["ExchangeItem"]()

    min_card = m.RoleTagAccessor("ownedMinCard")
    max_card = m.RoleTagAccessor("ownedMaxCard")


@m.xtype_handler(None)
class ExchangeItem(m.ModelElement):
    """An item that can be exchanged on an Exchange."""

    _xmltag = "ownedExchangeItems"

    type = m.EnumPOD(
        "exchangeMechanism", modeltypes.ExchangeMechanism, default="UNSET"
    )
    elements = m.DirectProxyAccessor(ExchangeItemElement, aslist=m.ElementList)
    exchanges: m.Accessor[m.ModelElement]


m.set_accessor(
    capellacore.Generalization, "super", m.AttrProxyAccessor(None, "super")
)
for cls in [Class, Union, datatype.Enumeration, Collection]:
    m.set_accessor(
        cls,
        "super",
        m.LinkAccessor(
            "ownedGeneralizations",
            capellacore.Generalization,
            attr="super",
            backattr="sub",
        ),
    )
    m.set_accessor(
        cls,
        "sub",
        m.ReferenceSearchingAccessor(cls, "super", aslist=m.MixedElementList),
    )

m.set_accessor(
    DataPkg, "packages", m.DirectProxyAccessor(DataPkg, aslist=m.ElementList)
)
m.set_accessor(
    Association,
    "members",
    m.RoleTagAccessor("ownedMembers", aslist=m.ElementList),
)
m.set_accessor(
    Association,
    "navigable_members",
    m.AttrProxyAccessor(Property, "navigableMembers", aslist=m.ElementList),
)
m.set_accessor(
    Class,
    "realized_classes",
    m.LinkAccessor[Class](
        "ownedInformationRealizations",
        InformationRealization,
        aslist=m.ElementList,
        attr="targetElement",
    ),
)
m.set_accessor(
    Class,
    "realizations",
    m.DirectProxyAccessor(InformationRealization, aslist=m.ElementList),
)
m.set_accessor(
    Class,
    "realized_by",
    m.ReferenceSearchingAccessor(
        Class, "realized_classes", aslist=m.ElementList
    ),
)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Implements a high-level interface to Capella projects."""

from __future__ import annotations

import collections.abc as cabc
import enum
import functools
import typing as t
import warnings

E = t.TypeVar("E", bound=enum.Enum)
"""TypeVar for ":py:class:`~enum.Enum`"."""
S = t.TypeVar("S", bound=str | None)
"""TypeVar for ":py:class:`str` | None"."""
T = t.TypeVar("T", bound="ModelObject")
"""TypeVar for ":py:class:`capellambse.model.ModelObject`"."""
T_co = t.TypeVar("T_co", bound="ModelObject", covariant=True)
"""Covariant TypeVar for ":py:class:`capellambse.model.ModelObject`"."""
U = t.TypeVar("U")
"""TypeVar (unbound)."""


def set_accessor(
    cls: type[ModelObject], attr: str, accessor: Accessor
) -> None:
    setattr(cls, attr, accessor)
    accessor.__set_name__(cls, attr)


def set_self_references(*args: tuple[type[ModelObject], str]) -> None:
    for cls, attr in args:
        set_accessor(cls, attr, DirectProxyAccessor(cls, aslist=ElementList))


def attr_equal(attr: str) -> cabc.Callable[[type[T]], type[T]]:
    def add_wrapped_eq(cls: type[T]) -> type[T]:
        orig_eq = cls.__eq__

        @functools.wraps(orig_eq)
        def new_eq(self, other: object) -> bool:
            try:
                cmpkey = getattr(self, attr)
            except AttributeError:
                pass
            else:
                if isinstance(other, type(cmpkey)):
                    result = cmpkey.__eq__(other)
                else:
                    result = other.__eq__(cmpkey)
                if result is not NotImplemented:
                    return result
            return orig_eq(self, other)

        cls.__eq__ = new_eq  # type: ignore[method-assign]
        cls.__hash__ = None  # type: ignore

        return cls

    return add_wrapped_eq


def stringy_enum(et: type[enum.Enum]) -> type[enum.Enum]:
    """Make an Enum stringy.

    This decorator makes an Enum's members compare equal to their
    respective ``name``.
    """

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self is other
        if isinstance(other, str):
            return self.name == other
        return NotImplemented

    def __str__(self):
        return str(self.name)

    def __hash__(self):
        return hash(self.name)

    et.__eq__ = __eq__  # type: ignore[method-assign]
    et.__str__ = __str__  # type: ignore[method-assign]
    et.__hash__ = __hash__  # type: ignore[method-assign]
    return et


from . import diagram
from ._descriptors import *
from ._model import *
from ._obj import *
from ._pods import *
from ._xtype import *

# NOTE: These are not in __all__ to avoid duplicate documentation in Sphinx,
#       however their docstring should mention the re-export.
from .diagram import AbstractDiagram as AbstractDiagram
from .diagram import Diagram as Diagram
from .diagram import DiagramAccessor as DiagramAccessor
from .diagram import DiagramType as DiagramType

ModelElement.parent = ParentAccessor(ModelElement)


if not t.TYPE_CHECKING:
    from ._descriptors import __all__ as _all1
    from ._model import __all__ as _all2
    from ._obj import __all__ as _all3
    from ._pods import __all__ as _all4
    from ._xtype import __all__ as _all5

    __all__ = [
        "E",
        "S",
        "T",
        "T_co",
        "U",
        "attr_equal",
        "diagram",
        "set_self_references",
        "stringy_enum",
        *_all1,
        *_all2,
        *_all3,
        *_all4,
        *_all5,
    ]

    _deprecated_names = {
        "GenericElement": ModelElement,
    }

    def __getattr__(attr):
        if target := _deprecated_names.get(attr):
            warnings.warn(
                f"{attr} is deprecated, use {target.__name__} instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return target
        raise AttributeError(f"{__name__} has no attribute {attr}")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Property Value Management extension for |project|."""

from __future__ import annotations

import logging
import typing as t

import capellambse.metamodel as mm
import capellambse.model as m

from ._config import *
from ._objects import *

LOGGER = logging.getLogger(__name__)


def _get_pvmt_configuration(model: m.MelodyModel) -> PVMTConfiguration:
    pkgs = model.project.property_value_packages
    assert pkgs.is_coupled()
    try:
        ext = pkgs.by_name("EXTENSIONS")
    except KeyError:
        LOGGER.debug("Creating EXTENSIONS package")
        ext = pkgs.create(name="EXTENSIONS")
    assert isinstance(ext, mm.capellacore.PropertyValuePkg)
    return PVMTConfiguration.from_model(model, ext._element)


def init() -> None:
    """Initialize the PVMT extension."""
    m.MelodyModel.pvmt = property(_get_pvmt_configuration)  # type: ignore[attr-defined]
    m.set_accessor(m.ModelElement, "pvmt", m.AlternateAccessor(ObjectPVMT))


if not t.TYPE_CHECKING:
    from ._config import __all__ as _all1
    from ._objects import __all__ as _all2

    __all__ = [*_all1, *_all2]



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Classes that allow access to diagrams in the model."""

from __future__ import annotations

__all__ = [
    # main interface
    "AbstractDiagram",
    "Diagram",
    "DiagramAccessor",
    "DiagramType",
    # format converters
    "DiagramFormat",
    "PrettyDiagramFormat",
    "convert_svgdiagram",
    "SVGFormat",
    "PNGFormat",
    "ConfluenceSVGFormat",
    "SVGDataURIFormat",
    "SVGInHTMLIMGFormat",
    "TerminalGraphicsFormat",
    # exceptions
    "UnknownOutputFormat",
    # helpers
    "convert_format",
    "REPR_DRAW",
]

import abc
import base64
import collections.abc as cabc
import contextlib
import enum
import importlib.metadata as imm
import io
import logging
import os
import traceback
import typing as t
import uuid

import markupsafe
from lxml import etree

import capellambse
from capellambse import aird, diagram, helpers, svg

from . import _descriptors, _obj, _pods, _xtype, stringy_enum


@stringy_enum
class DiagramType(enum.Enum):
    """The types of diagrams that Capella knows about.

    Extracted from::

        $CAPELLA/eclipse/configuration/org.eclipse.osgi/635/0/.cp/description

    with::

        grep '<ownedRepresentations' *(.) \
        | grep --color=always -P '(?<=name=").*?(?=")'

    Also directly exposed as ``capellambse.model.DiagramAccessor``.
    """

    UNKNOWN = "(Unknown Diagram Type)"
    # Common
    MSM = "Mode State Machine"
    # Capella Architecture?
    CDI = "Contextual Component Detailed Interfaces"
    CEI = "Contextual Component External Interfaces"
    CII = "Contextual Component Internal Interfaces"
    IDB = "Interfaces Diagram Blank"
    # Requirements?
    CRI = "Contextual Capability Realization Involvement"
    CRB = "Capability Realization Blank"
    PD = "Package Dependencies"
    ID = "Interface Delegations"
    CDB = "Class Diagram Blank"
    IS = "Component Interfaces Scenario"
    ES = "Component Exchanges Scenario"
    FS = "Functional Scenario"
    SFCD = LFCD = PFCD = "Functional Chain Description"
    # State And Mode - Matrix?
    # Contextual State And Mode - Matrix?
    # Modes and States Reference Matrix?
    # Operational Analysis
    # Operational Activities - Requirements?
    OEBD = "Operational Entity Breakdown"
    OAIB = "Operational Activity Interaction Blank"
    OAB = "Operational Entity Blank"
    OABD = "Operational Activity Breakdown"
    ORB = "Operational Role Blank"
    OES = "Operational Interaction Scenario"
    OAS = "Activity Interaction Scenario"
    OPD = "Operational Process Description"
    OCB = "Operational Capabilities Blank"
    # Requirements - Operational Activities?
    COC = "Contextual Operational Capability"
    # System Analysis
    CM = "Contextual Mission"
    MB = "Missions Blank"
    CC = "Contextual Capability"
    MCB = "Missions Capabilities Blank"
    # System Functions - Requirements?
    # System Functions - Operational Activities?
    SFBD = "System Function Breakdown"
    SDFB = "System Data Flow Blank"
    SAB = "System Architecture Blank"
    CSA = "Contextual System Actors"
    # System Actor - Operational Actor?
    # Interfaces - Capabilities?
    # Interfaces - Scenarios?
    # Interfaces - Capabilities and Scenarios?
    # System/Actors - System Functions?
    # Requirements - System Functions?
    # Logical Architecture
    # Logical Functions - Requirements?
    # Logical Components - Requirements?
    # Logical Functions - System Functions?
    # Logical Components - Logical Functions?
    # Logical Architecture Requirement Refinements?
    # Logical Interface - Context Interface?
    # Logical Actor - Context Actor?
    LCBD = "Logical Component Breakdown"
    LFBD = "Logical Function Breakdown"
    LDFB = "Logical Data Flow Blank"
    LAB = "Logical Architecture Blank"
    CRR = "Capability Realization Refinement"
    # Requirements - Logical Functions?
    # Physical Architecture
    # Physical Functions - Requirements?
    # Physical Components - Requirements?
    # Physical Functions - Logical Functions?
    # Physical Components - Logical Components?
    # Physical Components - Physical Functions?
    # Physical Interface - Logical Interface?
    PFBD = "Physical Function Breakdown"
    PDFB = "Physical Data Flow Blank"
    PCBD = "Physical Component Breakdown"
    PAB = "Physical Architecture Blank"
    # Physical Actor - Logical Actor?
    # Requirements - Physical Functions?
    PPD = "Physical Path Description"
    # EPBS
    # Configuration Items - Requirements?
    # Configuration Items - Physical Artifacts?
    # EPBS Requirement Refinements?
    EAB = "EPBS Architecture Blank"
    CIBD = "Configuration Items Breakdown"


@t.runtime_checkable
class DiagramFormat(t.Protocol):
    @classmethod
    def convert(cls, dg: t.Any) -> t.Any: ...


@t.runtime_checkable
class PrettyDiagramFormat(DiagramFormat, t.Protocol):
    @classmethod
    def convert_pretty(cls, dg: t.Any) -> t.Any: ...


DiagramConverter = t.Callable[[diagram.Diagram], t.Any] | DiagramFormat

LOGGER = logging.getLogger(__name__)
REPR_DRAW: bool
"""Whether to draw diagrams onto the terminal in their repr().

Determined automatically based on whether output is connected to a
terminal, and whether the terminal supports the kitty graphics protocol.
"""


class UnknownOutputFormat(ValueError):
    """An unknown output format was requested for the diagram."""


class AbstractDiagram(metaclass=abc.ABCMeta):
    """Abstract superclass of model diagrams.

    Also directly exposed as ``capellambse.model.AbstractDiagram``.
    """

    if t.TYPE_CHECKING:

        @property
        def uuid(self) -> str: ...
        @property
        def name(self) -> str: ...
        @property
        def target(self) -> _obj.ModelElement: ...

    else:
        uuid: str
        """Unique ID of this diagram."""
        name: str
        """Human-readable name for this diagram."""
        target: _obj.ModelElement
        """This diagram's "target".

        The target of a diagram is usually:

        *   The model element which is the direct parent of all visible
            nodes **OR**
        *   The only top-level element in the diagram **OR**
        *   The element which is considered to be the "element of interest".
        """

    filters: cabc.MutableSet[str]
    """The filters that are activated for this diagram."""

    if t.TYPE_CHECKING:

        @property
        def _allow_render(self) -> bool: ...

    else:
        _allow_render: bool = True
        """Allow this diagram to be rendered by the internal rendering engine.

        If this property is set to False, and a diagram cache was
        specified for the model, this diagram can only be loaded from
        the cache, and will never be rendered. Has no effect if there
        was no diagram cache specified.

        :meta public:
        """
    _model: capellambse.MelodyModel
    _render: diagram.Diagram
    _error: BaseException
    _last_render_params: dict[str, t.Any]
    """Additional rendering parameters for the cached rendered diagram.

    Rendering options for :class:`aird.Diagram`s. Handing over
    parameters that differ to these will force a fresh rendering of the
    diagram, flushing the cache.
    """

    def __init__(self, model: capellambse.MelodyModel) -> None:
        self._model = model
        self._last_render_params = {}

    def __dir__(self) -> list[str]:
        return dir(type(self)) + [
            f"as_{i.name}"
            for i in imm.entry_points()["capellambse.diagram.formats"]
        ]

    def __getattr__(self, attr: str) -> t.Any:
        if attr.startswith("as_"):
            fmt = attr[len("as_") :]
            try:
                return self.render(fmt)
            except UnknownOutputFormat:
                raise
            except Exception as err:
                if hasattr(self, "_error") and err is self._error:
                    err_img = self._render
                else:
                    err_img = self.__create_error_image("render", err)
                assert err_img is not None
                converter = _find_format_converter(fmt)
                data: t.Any = err_img
                for cv in reversed(list(_walk_converters(converter))):
                    if isinstance(cv, DiagramFormat):
                        data = cv.convert(data)
                    else:
                        data = cv(data)
                return data
        return getattr(super(), attr)

    def __repr__(self) -> str:
        if not __getattr__("REPR_DRAW"):  # type: ignore[name-defined]
            return self._short_repr_()

        try:
            escapes = self.render("termgraphics")
        except Exception:
            LOGGER.exception("Cannot render diagram as PNG for display")
            return self._short_repr_()

        return self._short_repr_() + "\n" + escapes.decode("ascii")

    def _short_repr_(self) -> str:
        return f"<Diagram {self.name!r}>"

    def __html__(self) -> markupsafe.Markup:
        return (
            markupsafe.Markup("<figure>")
            + markupsafe.Markup(self.render("svg"))
            + markupsafe.Markup("<figcaption>")
            + self.name
            + markupsafe.Markup("</figcaption></figure>")
        )

    def _short_html_(self) -> markupsafe.Markup:
        return markupsafe.Markup(
            f"<b>{markupsafe.Markup.escape(self.name)}</b> "
            f"(uuid: {markupsafe.Markup.escape(self.uuid)})"
        )

    def _repr_mimebundle_(
        self,
        include: cabc.Container[str] | None = None,
        exclude: cabc.Container[str] | None = None,
        **_kw,
    ) -> tuple[dict[str, t.Any], dict[t.Any, t.Any]] | dict[str, t.Any] | None:
        if include is None:
            include = helpers.EverythingContainer()
        if exclude is None:
            exclude = ()

        formats: dict[str, DiagramConverter] = {}
        for _, conv in _iter_format_converters():
            mime = getattr(conv, "mimetype", None)
            if not mime or mime not in include or mime in exclude:
                continue

            formats[mime] = conv
        if not formats:
            return None

        bundle: dict[str, t.Any] = {}
        for mime, conv in formats.items():
            with contextlib.suppress(KeyError):
                bundle[mime] = self.__load_cache([conv])

        if bundle:
            return bundle

        render = self.__render_fresh({})
        for mime, conv in formats.items():
            try:
                chain = list(_walk_converters(conv))
                bundle[mime] = _run_converter_chain(chain, render)
            except Exception:
                LOGGER.exception("Failed converting diagram with %r", conv)
        if not bundle:
            LOGGER.error("Failed converting diagram for MIME bundle")
            bundle["text/plain"] = repr(self)
        return bundle

    @property
    def nodes(self) -> _obj.MixedElementList:
        """Return a list of all nodes visible in this diagram."""
        allids = {e.uuid for e in self.render(None) if not e.hidden}
        elems = []
        for elemid in allids:
            assert elemid is not None
            try:
                elem = self._model.by_uuid(elemid)
            except KeyError:
                continue

            elems.append(elem._element)
        return _obj.MixedElementList(self._model, elems, _obj.ModelElement)

    @t.overload
    def render(self, fmt: None, /, **params) -> diagram.Diagram: ...
    @t.overload
    def render(
        self, fmt: str, /, *, pretty_print: bool = ..., **params
    ) -> t.Any: ...
    def render(
        self,
        fmt: str | None,
        /,
        *,
        pretty_print: bool = False,
        **params,
    ) -> t.Any:
        """Render the diagram in the given format.

        Parameters
        ----------
        fmt
            The output format to use.

            If ``None``, the :class:`~capellambse.diagram.Diagram` is returned
            without format conversion.
        pretty_print
            Whether to pretty-print the output. Only applies to
            text-based formats. Ignored if the output format converter
            does not support pretty-printing.
        params
            Additional render parameters. Which parameters are
            supported depends on the specific type of diagram.
        """
        if fmt is None:
            chain: list[DiagramConverter] = [lambda i: i]
        else:
            chain = list(_walk_converters(_find_format_converter(fmt)))
            if self._model.diagram_cache is not None:
                try:
                    return self.__load_cache(chain)
                except KeyError:
                    pass

                if not self._allow_render:
                    raise RuntimeError(f"Diagram not in cache: {self.name}")

        render = self.__render_fresh(params)
        return _run_converter_chain(chain, render, pretty_print=pretty_print)

    def save(
        self,
        file: str | os.PathLike | t.IO[bytes] | None,
        fmt: str,
        /,
        *,
        pretty_print: bool = False,
        **params,
    ) -> None:
        """Render the diagram and write it to a file.

        Parameters
        ----------
        file
            The file to write the diagram to. Can be a filename, or a
            file-like object in binary mode.

            Text-based formats that render to a :class:`str` will always
            be encoded as UTF-8.

            If None is passed, and the selected format has a known
            filename extension, a filename will be generated from the
            diagram's name and the extension.
        fmt
            The output format to use.
        pretty_print
            Whether to pretty-print the output. Only applies to
            text-based formats. Ignored if the output format converter
            does not support pretty-printing.
        params
            Additional render parameters to pass to the :meth:`render`
            call.
        """
        if file is None:
            conv = _find_format_converter(fmt)
            if ext := getattr(conv, "filename_extension", None):
                file = f"{self.name} ({self.uuid}){ext}"
            else:
                raise ValueError(
                    f"No known extension for format {fmt!r},"
                    " specify a file name explicitly"
                )
        data = self.render(fmt, pretty_print=pretty_print, **params)
        if isinstance(data, str):
            data = data.encode("utf-8")
        elif not isinstance(data, bytes):
            raise TypeError(
                f"Cannot write format {fmt!r} to file:"
                f" expected str or bytes, got {type(data).__name__}"
            )
        if hasattr(file, "write"):
            file.write(data)
        else:
            with open(file, "wb") as f:
                f.write(data)

    @abc.abstractmethod
    def _create_diagram(self, params: dict[str, t.Any]) -> diagram.Diagram:
        """Perform the actual rendering of the diagram.

        This method is called by :meth:`.render` to perform the actual
        rendering of the diagram. It is passed the parameters that were
        passed to :meth:`.render` as a dictionary.

        Subclasses override this method to implement their rendering
        logic. Do not call this method directly, use :meth:`.render`
        instead - it will take care of caching and properly converting
        the render output.

        :meta public:
        """

    def __create_error_image(
        self, stage: str, error: Exception
    ) -> diagram.Diagram:
        err_name = (
            "An error occured while rendering diagram\n"
            f"{self.name!r}\n"
            f"(in stage {stage!r})"
        )
        err_msg = (
            "Please report this error to your tools and methods team,"
            " and attach the following information:"
        )
        err_trace = "".join(
            traceback.format_exception(None, error, error.__traceback__)
        )

        diag = diagram.Diagram("An error occured! :(")
        err_box = diagram.Box(
            (200, 0),
            (350, 0),
            label=err_name,
            uuid="error",
            styleclass="Note",
            styleoverrides={
                "fill": diagram.RGB(255, 0, 0),
                "text_fill": diagram.RGB(255, 255, 255),
            },
        )
        diag.add_element(err_box, extend_viewport=False)
        info_box = diagram.Box(
            (200, err_box.pos.y + err_box.size.y + 10),
            (350, 0),
            label=err_msg,
            uuid="info",
            styleclass="Note",
        )
        diag.add_element(info_box, extend_viewport=False)
        trace_box = diagram.Box(
            (0, info_box.pos.y + info_box.size.y + 10),
            (750, 0),
            label=err_trace,
            uuid="trace",
            styleclass="Note",
        )
        diag.add_element(trace_box, extend_viewport=False)
        diag.calculate_viewport()
        return diag

    def __load_cache(self, chain: list[DiagramConverter]) -> t.Any:
        cache_handler = self._model.diagram_cache
        if cache_handler is None:
            raise KeyError(self.uuid)

        data: t.Any = None
        for i, cv in enumerate(chain):
            ext = getattr(cv, "filename_extension", None)
            if not ext or not hasattr(cv, "from_cache"):
                continue

            filename = self.uuid + ext
            try:
                with cache_handler.open(filename) as f:
                    cache = f.read()
            except FileNotFoundError:
                LOGGER.debug("Not found in diagram cache: %s", filename)
            else:
                LOGGER.debug("Using file from diagram cache: %s", filename)
                data = cv.from_cache(cache)
                chain = chain[:i]
                break
        else:
            LOGGER.debug(
                "No usable cached format found for diagram %s (%r)",
                self.uuid,
                self.name,
            )
            raise KeyError(self.uuid)

        for cv in reversed(chain):
            LOGGER.debug("Executing format converter %r", cv)
            if hasattr(cv, "convert"):
                data = cv.convert(data)
            else:
                data = cv(data)
        return data

    def __render_fresh(self, params: dict[str, t.Any]) -> diagram.Diagram:
        if not hasattr(self, "_render") or self._last_render_params != params:
            self.invalidate_cache()
            try:
                self._render = self._create_diagram(params)
            except Exception as err:
                self._error = err
                self._render = self.__create_error_image("parse", self._error)

        if hasattr(self, "_error"):
            raise self._error
        return self._render

    def invalidate_cache(self) -> None:
        """Reset internal diagram cache."""
        with contextlib.suppress(AttributeError):
            del self._render

        with contextlib.suppress(AttributeError):
            del self._error


@_xtype.xtype_handler(
    None,
    "viewpoint:DRepresentationDescriptor",
    "diagram:DSemanticDiagram",
    "sequence:SequenceDDiagram",
)
class Diagram(AbstractDiagram):
    """Provides access to a single diagram.

    Also directly exposed as ``capellambse.model.Diagram``.
    """

    uuid: str = _pods.StringPOD(  # type: ignore[assignment]
        "uid", writable=False
    )
    xtype = property(lambda self: helpers.xtype_of(self._element))
    name: str = _pods.StringPOD("name")  # type: ignore[assignment]
    description = _pods.HTMLStringPOD("documentation")

    _element: aird.DRepresentationDescriptor

    __nodes: list[etree._Element] | None

    @classmethod
    def from_model(
        cls,
        model: capellambse.model.MelodyModel,
        element: etree._Element,
    ) -> Diagram:
        """Wrap a diagram already defined in the Capella AIRD."""
        if aird.is_representation_descriptor(element):
            self = cls.__new__(cls)
            self._model = model
            self._element = element
            self._last_render_params = {}
            self.__nodes = None
            return self

        target_id = element.get("uid")
        if not target_id:
            raise RuntimeError(f"No uid defined on {element!r}")
        return model.diagrams.by_representation_path(
            f"#{target_id}", single=True
        )

    def __init__(self, **kw: t.Any) -> None:
        del kw
        raise TypeError("Cannot create a Diagram this way")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._model is other._model and self._element == other._element

    @property
    def nodes(self) -> _obj.MixedElementList:
        if self.__nodes is None:
            self.__nodes = list(
                aird.iter_visible(self._model._loader, self._element)
            )
        return _obj.MixedElementList(
            self._model, self.__nodes.copy(), _obj.ModelElement
        )

    @property
    def semantic_nodes(self) -> _obj.MixedElementList:
        if self.__nodes is None:
            self.__nodes = list(
                aird.iter_visible(self._model._loader, self._element)
            )

        from capellambse.metamodel import cs, interaction

        elems: list[etree._Element] = []
        for i in self.__nodes:
            obj = _obj.ModelElement.from_model(self._model, i)
            if isinstance(obj, cs.Part):
                obj = obj.type
            elif isinstance(obj, interaction.AbstractInvolvement):
                obj = obj.involved
            elif isinstance(obj, interaction.StateFragment):
                obj = obj.function

            if obj is not None:
                elems.append(obj._element)
        return _obj.MixedElementList(self._model, elems)

    @property
    def _allow_render(self) -> bool:
        return self._model._fallback_render_aird

    @property
    def viewpoint(self) -> str:
        return aird.viewpoint_of(self._element)

    @property
    def representation_path(self) -> str:
        """The ID of the representation, i.e. the actual diagram.

        Capella distinguishes between the representation (which contains
        all visual data) and a representation descriptor, which only
        contains a handful of metadata like the diagram name and
        description. Both have their own IDs, and the descriptor
        contains a link to the representation.
        """
        return self._element.attrib["repPath"]

    @property
    def target(self) -> _obj.ModelElement:
        target = aird.find_target(self._model._loader, self._element)
        return _obj.ModelElement.from_model(self._model, target)

    @property
    def type(self) -> DiagramType:
        """Return the type of this diagram."""
        sc = aird.get_styleclass(self._element)
        try:
            return DiagramType(sc)
        except ValueError:  # pragma: no cover
            LOGGER.warning("Unknown diagram type %r", sc)
            return DiagramType.UNKNOWN

    @property
    def filters(self) -> cabc.MutableSet[str]:
        """Return a set of currently activated filters on this diagram."""
        return aird.ActiveFilters(self._model, self)

    @filters.setter
    def filters(self, filters: cabc.Iterable[str]) -> None:
        self.filters.clear()
        for filter in filters:
            self.filters.add(filter)

    def _create_diagram(self, params: dict[str, t.Any]) -> diagram.Diagram:
        return aird.parse_diagram(self._model._loader, self._element, **params)

    def invalidate_cache(self) -> None:
        """Reset internal diagram cache."""
        self.__nodes = None
        super().invalidate_cache()


class DiagramAccessor(_descriptors.Accessor):
    """Provides access to a list of diagrams below the specified viewpoint.

    Also directly exposed as ``capellambse.model.DiagramAccessor``.
    """

    def __init__(
        self,
        viewpoint: str | None = None,
        *,
        cacheattr: str | None = None,
    ) -> None:
        super().__init__()
        self.cacheattr = cacheattr
        self.viewpoint = viewpoint

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        if isinstance(obj, _model.MelodyModel):
            model = obj
        else:
            model = obj._model
        descriptors = aird.enumerate_descriptors(
            model._loader, viewpoint=self.viewpoint
        )
        return _obj.CachedElementList(
            model,
            list(descriptors),
            Diagram,
            cacheattr=self.cacheattr,
        )


def convert_svgdiagram(
    dg: diagram.Diagram,
) -> svg.generate.SVGDiagram:
    """Convert the diagram to a SVGDiagram."""
    jsondata = diagram.DiagramJSONEncoder().encode(dg)
    return svg.generate.SVGDiagram.from_json(jsondata)


class SVGFormat:
    """Convert the diagram to SVG."""

    filename_extension = ".svg"
    mimetype = "image/svg+xml"
    depends = convert_svgdiagram

    @staticmethod
    def convert(dg: svg.generate.SVGDiagram) -> str:
        return dg.to_string()

    @staticmethod
    def convert_pretty(dg: svg.generate.SVGDiagram) -> str:
        buf = io.StringIO()
        drawing = dg.drawing._Drawing__drawing  # type: ignore[attr-defined]
        drawing.write(buf, pretty=True)
        return buf.getvalue()

    @staticmethod
    def from_cache(cache: bytes) -> str:
        return cache.decode("utf-8")


class PNGFormat:
    """Convert the diagram to PNG."""

    filename_extension = ".png"
    mimetype = "image/png"
    depends = SVGFormat

    @staticmethod
    def convert(dg: str) -> bytes:
        try:
            import cairosvg
        except OSError as error:
            raise RuntimeError(
                "Cannot import cairosvg. You are likely missing .dll's."
                " Please see the README for instructions."
            ) from error

        scale = float(os.getenv("CAPELLAMBSE_PNG_SCALE", "1"))
        return cairosvg.svg2png(dg, scale=scale)

    @staticmethod
    def from_cache(cache: bytes) -> bytes:
        return cache


class ConfluenceSVGFormat:
    """Convert the diagram to Confluence-style SVG."""

    depends = SVGFormat

    prefix = (
        f'<ac:structured-macro ac:macro-id="{uuid.uuid4()!s}" ac:name="html"'
        ' ac:schema-version="1">'
        "<ac:plain-text-body><![CDATA["
    )
    postfix = "]]></ac:plain-text-body></ac:structured-macro>"

    @classmethod
    def convert(cls, dg: str) -> str:
        return "".join((cls.prefix, dg, cls.postfix))

    @classmethod
    def convert_pretty(cls, dg: str) -> str:
        return cls.prefix + "\n" + dg + "\n" + cls.postfix


class SVGDataURIFormat:
    depends = SVGFormat

    preamble = "data:image/svg+xml;base64,"

    @classmethod
    def convert(cls, dg: str) -> str:
        b64 = base64.standard_b64encode(dg.encode("utf-8"))
        return "".join((cls.preamble, b64.decode("ascii")))


class SVGInHTMLIMGFormat:
    depends = SVGDataURIFormat

    @staticmethod
    def convert(dg: str) -> markupsafe.Markup:
        return markupsafe.Markup(f'<img src="{dg}"/>')


class TerminalGraphicsFormat:
    """The kitty terminal graphics protocol diagram format.

    This graphics format generates terminal escape codes that transfer
    PNG data to a TTY using the `kitty graphics protocol`__.

    __ https://sw.kovidgoyal.net/kitty/graphics-protocol/
    """

    depends = PNGFormat

    @classmethod
    def convert(cls, dg: bytes) -> bytes:
        container = b"\x1b_Ga=T,q=2,f=100,m=%d;%b\x1b\\"

        chunks: list[bytes] = []
        png_b64 = base64.standard_b64encode(dg)
        while png_b64:
            chunk, png_b64 = png_b64[:4096], png_b64[4096:]
            m = (0, 1)[bool(png_b64)]
            chunks.append(container % (m, chunk))

        return b"".join(chunks)

    @staticmethod
    def is_supported() -> bool:
        """Return whether the used terminal supports graphics.

        This implementation checks whether stdin, stdout and stderr are
        connected to a terminal, and whether the ``$TERM`` environment
        variable is set to a know-supportive value. Currently the only
        recognized value is ``xterm-kitty``.
        """
        return (
            all(os.isatty(i) for i in range(3))
            and os.getenv("TERM") == "xterm-kitty"
        )


def convert_format(
    sourcefmt: str | None,
    targetfmt: str,
    data: t.Any,
    /,
    *,
    pretty_print: bool = False,
) -> t.Any:
    """Convert between different image formats.

    Parameters
    ----------
    sourcefmt
        Name of the current format. If None, the current format is a
        :class:`capellambse.diagram.Diagram` object.
    targetfmt
        Name of the target format.
    data
        Image data in the format produced by the "sourcefmt" converter.
    pretty_print
        Whether to instruct the converters to pretty-print their output.
        Only relevant for text-based formats like SVG.

    Returns
    -------
    ~typing.Any
        The converted image data.

    Raises
    ------
    ValueError
        Raised if either format is not known, or if no converter chain
        could be built to convert between the two formats.
    """
    if sourcefmt is None:
        source = None
    else:
        source = _find_format_converter(sourcefmt)
    chain: list[DiagramConverter] = []
    for i in _walk_converters(_find_format_converter(targetfmt)):
        if i is source:
            break
        chain.append(i)
    else:
        raise ValueError(f"Cannot convert from {sourcefmt} to {targetfmt}")

    return _run_converter_chain(chain, data, pretty_print=pretty_print)


def _run_converter_chain(
    chain: list[DiagramConverter],
    data: t.Any,
    pretty_print: bool = False,
) -> t.Any:
    for conv in reversed(chain):
        LOGGER.debug("Executing format converter %r", conv)
        if pretty_print and isinstance(conv, PrettyDiagramFormat):
            data = conv.convert_pretty(data)
        elif isinstance(conv, DiagramFormat):
            data = conv.convert(data)
        else:
            data = conv(data)
    return data


def _find_format_converter(fmt: str) -> DiagramConverter:
    eps = imm.entry_points(group="capellambse.diagram.formats", name=fmt)
    if not eps:
        raise UnknownOutputFormat(f"Unknown image output format {fmt}")
    return next(iter(eps)).load()


def _iter_format_converters() -> t.Iterator[tuple[str, DiagramConverter]]:
    for ep in imm.entry_points(group="capellambse.diagram.formats"):
        try:
            conv = ep.load()
        except ImportError:
            pass
        else:
            yield (ep.name, conv)


def _walk_converters(
    first: DiagramConverter,
) -> cabc.Iterator[DiagramConverter]:
    yield first
    current: DiagramConverter | None = first
    while (current := getattr(current, "depends", None)) is not None:
        yield current


if not t.TYPE_CHECKING:

    def __getattr__(name):
        if name != "REPR_DRAW":
            raise AttributeError(f"No name {name} in module {__name__}")
        try:
            return globals()["REPR_DRAW"]
        except KeyError:
            pass
        globals()["REPR_DRAW"] = d = TerminalGraphicsFormat.is_supported()
        return d


from . import _model



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import pathlib
import shutil

import pytest

import capellambse
from capellambse.extensions import pvmt

TEST_ROOT = pathlib.Path(__file__).parent / "data" / "pvmt"
MODEL_FILE = "PVMTTest.aird"


@pytest.fixture
def model(monkeypatch, tmp_path):
    new_test_root = tmp_path / "model"
    shutil.copytree(TEST_ROOT, new_test_root)
    monkeypatch.setitem(globals(), "TEST_ROOT", new_test_root)
    return capellambse.MelodyModel(new_test_root)


class TestPVMTConfiguration:
    def test_domains(self, model):
        expected = {
            "02e0c435-f085-471f-9f6e-e12fe5f27687": "Computer",
            "12a02f1b-8b97-4188-95ea-2afd377c4c41": "Out of scope",
            "ae15f861-2f15-40d2-a1b9-57dfdf7fabae": "In scope",
            "c2d3e3a3-f9bb-4ff7-964d-2109c7065bc2": "External Data",
        }

        actual = {i.uuid: i.name for i in model.pvmt.domains}

        assert actual == expected

    @pytest.mark.parametrize(
        ("enum_uuid", "enum_vals"),
        [
            (
                "03e5e5ae-0a61-473d-a792-e003ce601ff4",
                {"UNSET", "Trace", "Cable"},
            ),
            (
                "ae279311-6a75-40b7-9bf3-a19c136d0573",
                {"UNSET", "Core Component", "External"},
            ),
        ],
    )
    def test_enums(self, model, enum_uuid, enum_vals):
        enum = model.by_uuid(enum_uuid)
        domain = model.pvmt.domains["Computer"]
        assert enum in domain.enumeration_property_types

        actual = {i.name for i in enum.literals}
        assert actual == enum_vals

    def test_groups(self, model):
        expected = {"Components", "Cables", "Physical Cables"}

        model_groups = model.pvmt.domains["Computer"].groups
        actual = {g.name for g in model_groups}

        assert actual == expected

    @pytest.mark.parametrize(
        "group",
        [
            "Architecture",
            "Property (Bool, not equal)",
            "Property (Enum, equal)",
            "Property (Enum, not equal)",
            "Property (Float, equal)",
            "Property (Float, not equal)",
            "Property (Float, greater than)",
            "Property (Float, less than)",
            "Property (Float, greater or equal)",
            "Property (Float, less or equal)",
            "Property (Integer, equal)",
            "Property (Integer, not equal)",
            "Property (Integer, greater than)",
            "Property (Integer, less than)",
            "Property (Integer, greater or equal)",
            "Property (Integer, less or equal)",
            "Property (String, equal)",
            "Property (String, not equal)",
            "Property (String, contains)",
            "Property (String, starts with)",
            "Property (String, ends with)",
        ],
    )
    def test_apply_outofscope(self, model, group):
        obj = model.by_uuid("d32caffc-b9a1-448e-8e96-65a36ba06292")
        domain = model.pvmt.domains["Out of scope"]
        group = domain.groups[group]
        with pytest.raises(pvmt.ScopeError):
            group.apply(obj)

    @pytest.mark.parametrize(
        "group",
        [
            "Property (Bool, equal)",
            "Property (String, equal)",
        ],
    )
    def test_apply_inscope(self, model, group):
        obj = model.by_uuid("d32caffc-b9a1-448e-8e96-65a36ba06292")
        domain = model.pvmt.domains["In scope"]
        group = domain.groups[group]

        ag = group.apply(obj)

        assert ag is not None
        assert ag.name == f"In scope.{group.name}"
        assert len(ag.property_values) == len(group.property_values)
        applied_values = {i.name: i.value for i in ag.property_values}
        default_values = {i.name: i.value for i in group.property_values}
        assert applied_values == default_values


class TestAppliedPropertyValueGroupXML:
    """Tests that rely on writing back and comparing the output XML."""

    expect_root = TEST_ROOT / "expected-output"

    elem_uuid = "d32caffc-b9a1-448e-8e96-65a36ba06292"

    def compare_xml(self, model, expected_file):
        pvmt_model_path = (TEST_ROOT / MODEL_FILE).with_suffix(".capella")
        expected_model_path = self.expect_root / expected_file

        model.save()
        actual = pvmt_model_path.read_text(encoding="utf-8")
        expected = expected_model_path.read_text(encoding="utf-8")

        assert actual == expected

    def test_apply(self, model, monkeypatch):
        call_count = 0

        def mock_generate_uuid(*__, **_):
            nonlocal call_count
            call_count += 1
            return f"00000000-0000-0000-0000-{call_count:012x}"

        monkeypatch.setattr(
            "capellambse.loader.MelodyLoader.generate_uuid",
            mock_generate_uuid,
        )

        elem = model.by_uuid(self.elem_uuid)

        elem.pvmt["External Data.Object IDs.Object ID"] = "CABLE-0001"

        assert call_count == 2
        self.compare_xml(model, "apply.capella")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Implements the Capella Filtering extension."""

from __future__ import annotations

import collections.abc as cabc
import contextlib
import logging
import operator
import pathlib
import shutil
import sys
import typing as t

import typing_extensions as te

import capellambse
import capellambse.metamodel as mm
import capellambse.model as m
from capellambse import _native, helpers

VIEWPOINT: t.Final = "org.polarsys.capella.filtering"
NAMESPACE: t.Final = "http://www.polarsys.org/capella/filtering/6.0.0"
SYMBOLIC_NAME: t.Final = "filtering"

_LOGGER = logging.getLogger(__name__)

m.XTYPE_ANCHORS[__name__] = SYMBOLIC_NAME


@m.xtype_handler(None)
class FilteringCriterion(m.ModelElement):
    """A single filtering criterion."""

    _xmltag = "ownedFilteringCriteria"

    filtered_objects = m.ReferenceSearchingAccessor[m.ModelElement](
        (), "filtering_criteria", aslist=m.MixedElementList
    )


@m.xtype_handler(None)
class FilteringCriterionPkg(m.ModelElement):
    """A package containing multiple filtering criteria."""

    _xmltag = "ownedFilteringCriterionPkgs"

    criteria = m.DirectProxyAccessor(FilteringCriterion, aslist=m.ElementList)
    packages: m.Accessor[FilteringCriterionPkg]


@m.xtype_handler(None)
class FilteringModel(m.ModelElement):
    """A filtering model containing criteria to filter by."""

    criteria = m.DirectProxyAccessor(FilteringCriterion, aslist=m.ElementList)
    criterion_packages = m.DirectProxyAccessor(
        FilteringCriterionPkg, aslist=m.ElementList
    )


class AssociatedCriteriaAccessor(m.PhysicalAccessor[FilteringCriterion]):
    def __init__(self) -> None:
        super().__init__(
            FilteringCriterion, aslist=m.ElementList[FilteringCriterion]
        )

    @t.overload
    def __get__(self, obj: None, objtype: type[t.Any]) -> te.Self: ...
    @t.overload
    def __get__(
        self, obj: m.ModelObject, objtype: type[t.Any] | None = ...
    ) -> m.ElementList[m.T]: ...
    def __get__(
        self,
        obj: m.ModelObject | None,
        objtype: type[t.Any] | None = None,
    ) -> te.Self | m.ElementList[m.T]:
        del objtype
        if obj is None:  # pragma: no cover
            return self

        loader = obj._model._loader
        try:
            xt_critset = f"{SYMBOLIC_NAME}:AssociatedFilteringCriterionSet"
            critset = next(loader.iterchildren_xt(obj._element, xt_critset))
        except StopIteration:
            elems = []
        else:
            links = critset.get("filteringCriteria", "")
            elems = list(loader.follow_links(obj._element, links))

        return self._make_list(obj, elems)


@m.xtype_handler(None)
class FilteringResult(m.ModelElement):
    """A filtering result."""


@m.xtype_handler(None)
class ComposedFilteringResult(m.ModelElement):
    """A composed filtering result."""


def init() -> None:
    m.set_accessor(
        mm.capellamodeller.SystemEngineering,
        "filtering_model",
        m.DirectProxyAccessor(FilteringModel),
    )
    m.MelodyModel.filtering_model = property(  # type: ignore[attr-defined]
        operator.attrgetter("project.model_root.filtering_model")
    )
    m.set_accessor(
        m.ModelElement, "filtering_criteria", AssociatedCriteriaAccessor()
    )


m.set_self_references(
    (FilteringCriterionPkg, "packages"),
)

try:
    import click

    from capellambse import cli_helpers
except ImportError:

    def _main() -> None:
        """Display a dependency error."""
        print("Missing dependency 'capellambse[cli]'", file=sys.stderr)
        raise SystemExit(1)

else:

    @click.command()
    @click.argument("command")
    @click.option(
        "-m",
        "--model",
        "model_",
        type=cli_helpers.ModelCLI(),
        required=True,
        help="Source Capella model to derive from",
    )
    @click.option(
        "-r",
        "--result",
        "result_strings",
        multiple=True,
        help="A filtering result object, specified either by UUID or name.",
    )
    @click.option(
        "-o",
        "--output",
        type=click.Path(
            file_okay=False, dir_okay=True, path_type=pathlib.Path
        ),
        default=".",
        help="Directory to store the derived project(s) in.",
    )
    @click.option(
        "-p",
        "--name-pattern",
        default="{model.name}_derived_{result.name}",
        help=(
            "The 'str.format()' pattern for generating output project names."
            " Available format arguments are 'model' and the current 'result'."
        ),
    )
    @click.option("--exe", help="Name or path of the Capella executable")
    @click.option("--docker", help="Name of a Docker image containing Capella")
    def _main(
        command: str,
        model_: capellambse.MelodyModel,
        result_strings: list[str],
        output: pathlib.Path,
        name_pattern: str,
        exe: str | None,
        docker: str | None,
    ) -> None:
        """Manage filtered Capella projects.

        Currently the only supported COMMAND is "derive", which performs
        project derivation based on an existing FilteringResult.

        The native Capella installation (options ``--exe`` or
        ``--docker``) may contain the placeholder '{VERSION}', which
        will be replaced by the Capella version number that the model
        was created with (e.g. "6.0.0").

        Exactly one of ``--exe`` or ``--docker`` must be specified.
        """
        if command != "derive":
            click.echo(f"Unsupported command: {command}", err=True)
            raise SystemExit(1)

        results = _find_results(model_, result_strings)
        output.mkdir(exist_ok=True, parents=True)
        with (
            _native.native_capella(model_, exe=exe, docker=docker) as cli,
            click.progressbar(
                results, item_show_func=lambda i: getattr(i, "name", "")
            ) as progress,
        ):
            for result in progress:
                derived_name = name_pattern.format(model=model_, result=result)
                output_sub = output / derived_name
                with contextlib.suppress(FileNotFoundError):
                    shutil.rmtree(output_sub)
                proc = cli(
                    *_native.ARGS_CMDLINE,
                    "-appid",
                    "org.polarsys.capella.filtering.commandline",
                    "-input",
                    "main_model",
                    "-filteringresultid",
                    result.uuid,
                    "-derivationprojectname",
                    derived_name,
                )
                derived = cli.workspace / derived_name
                if not derived.exists():
                    _LOGGER.error(
                        "Failed to derive result %r (%s)\n\n%s",
                        result.name,
                        result.uuid,
                        proc.stdout,
                    )
                    raise RuntimeError(
                        "Output directory does not exist for derived result"
                        f" {result.name!r} ({result.uuid})"
                    )
                shutil.copytree(derived, output_sub)

    def _find_results(
        loaded_model: capellambse.MelodyModel, result_strings: list[str]
    ) -> cabc.Sequence[FilteringResult | ComposedFilteringResult]:
        all_results: m.ElementList[FilteringResult | ComposedFilteringResult]
        all_results = loaded_model.search(
            FilteringResult, ComposedFilteringResult
        )
        if not result_strings:
            return all_results

        wanted: list[FilteringResult | ComposedFilteringResult] = []
        obj: m.ModelElement
        for result in result_strings:
            if helpers.is_uuid_string(result):
                try:
                    obj = loaded_model.by_uuid(result)
                except KeyError:
                    pass
                else:
                    if not isinstance(
                        obj, FilteringResult | ComposedFilteringResult
                    ):
                        click.echo(
                            f"Error: Object {obj._short_repr_()} is of type"
                            f" {type(obj).__name__}, expected"
                            f" a FilteringResult or ComposedFilteringResult",
                            err=True,
                        )
                        raise SystemExit(1)
                    wanted.append(obj)
                    continue

            obj = all_results.by_name(result, single=True)
            assert isinstance(obj, FilteringResult | ComposedFilteringResult)
            wanted.append(obj)
        return wanted


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""The 'Requirements' namespace."""

from __future__ import annotations

__all__ = [
    "AbstractRequirementsAttribute",
    "AbstractRequirementsRelation",
    "AbstractType",
    "AttributeAccessor",
    "AttributeDefinition",
    "AttributeDefinitionEnumeration",
    "BooleanValueAttribute",
    "DataTypeDefinition",
    "DateValueAttribute",
    "EnumValue",
    "EnumerationDataTypeDefinition",
    "EnumerationValueAttribute",
    "Folder",
    "IntegerValueAttribute",
    "InternalRelation",
    "ModuleType",
    "RealValueAttribute",
    "RelationType",
    "ReqIFElement",
    "Requirement",
    "RequirementType",
    "StringValueAttribute",
]

import re
import typing as t

import capellambse.model as m
from capellambse import helpers

if t.TYPE_CHECKING:
    import markupsafe

m.XTYPE_ANCHORS[__name__] = "Requirements"


class ReqIFElement(m.ModelElement):
    """Attributes shared by all ReqIF elements."""

    identifier = m.StringPOD("ReqIFIdentifier")
    long_name = m.StringPOD("ReqIFLongName")
    description: str = m.StringPOD(  # type: ignore[assignment]
        "ReqIFDescription"
    )
    name = m.StringPOD("ReqIFName")
    prefix = m.StringPOD("ReqIFPrefix")
    type: m.Accessor = property(lambda _: None)  # type: ignore[assignment]

    def _short_repr_(self) -> str:  # pragma: no cover
        mytype = type(self).__name__
        parent = self._element
        if self.xtype in {  # FIXME remove this shit
            "CapellaRequirements:CapellaTypesFolder",
            "Requirements:DataTypeDefinition",
            "Requirements:RequirementType",
            "Requirements:RelationType",
            "Requirements:ModuleType",
            "Requirements:EnumerationDataTypeDefinition",
            "Requirements:EnumValue",
            "Requirements:AttributeDefinition",
            "Requirements:AttributeDefinitionEnumeration",
        }:
            return f'<{mytype} {parent.get("ReqIFLongName")!r} ({self.uuid})>'

        name = (
            parent.get("ReqIFName")
            or parent.get("ReqIFChapterName")
            or parent.get("ReqIFLongName")
            or ""
        )
        return f"<{mytype} {name!r} ({self.uuid})>"

    def _short_html_(self) -> markupsafe.Markup:
        name = self.name or self.long_name
        return helpers.make_short_html(type(self).__name__, self.uuid, name)


@m.xtype_handler(None)
class DataTypeDefinition(ReqIFElement):
    """A data type definition for requirement types."""

    _xmltag = "ownedDefinitionTypes"


@m.xtype_handler(None)
class AttributeDefinition(ReqIFElement):
    """An attribute definition for requirement types."""

    _xmltag = "ownedAttributes"

    data_type = m.AttrProxyAccessor(DataTypeDefinition, "definitionType")


class AbstractRequirementsAttribute(m.ModelElement):
    _xmltag = "ownedAttributes"

    definition = m.AttrProxyAccessor(AttributeDefinition, "definition")

    value: t.Any

    def __repr__(self) -> str:
        return self._short_repr_()

    def _short_repr_(self) -> str:
        mytype = type(self).__name__
        if self.definition is not None:
            defname = self.definition.long_name
        else:
            defname = re.sub(r"(?<!^)([A-Z])", r" \1", mytype)
        return f"<{mytype} [{defname}] {self._repr_value()} ({self.uuid})>"

    def _short_html_(self) -> markupsafe.Markup:
        if self.definition is not None:
            name = self.definition.long_name
        else:
            name = "None"
        return helpers.make_short_html(type(self).__name__, self.uuid, name)

    def _repr_value(self) -> str:
        return repr(self.value)


class AttributeAccessor(m.DirectProxyAccessor[AbstractRequirementsAttribute]):
    def __init__(self) -> None:
        super().__init__(
            m.ModelElement,  # type: ignore[arg-type]
            (
                BooleanValueAttribute,
                DateValueAttribute,
                EnumerationValueAttribute,
                IntegerValueAttribute,
                RealValueAttribute,
                StringValueAttribute,
            ),
            aslist=m.MixedElementList,
            mapkey="definition.long_name",
            mapvalue="value",
        )

    def _match_xtype(
        self, type_: str, /
    ) -> tuple[type[AbstractRequirementsAttribute], str]:
        type_ = type_.lower()
        try:
            cls = _attr_type_hints[type_]
            return cls, m.build_xtype(cls)
        except KeyError:
            raise ValueError(f"Invalid type hint given: {type_!r}") from None


@m.xtype_handler(None)
class BooleanValueAttribute(AbstractRequirementsAttribute):
    """A string value attribute."""

    value = m.BoolPOD("value")


@m.xtype_handler(None)
class DateValueAttribute(AbstractRequirementsAttribute):
    """A value attribute that stores a date and time."""

    value = m.DatetimePOD("value")

    def _repr_value(self) -> str:
        if self.value is None:
            return "None"
        return self.value.isoformat()


@m.xtype_handler(None)
class IntegerValueAttribute(AbstractRequirementsAttribute):
    """An integer value attribute."""

    value = m.IntPOD("value")


@m.xtype_handler(None)
class RealValueAttribute(AbstractRequirementsAttribute):
    """A floating-point number value attribute."""

    value = m.FloatPOD("value")


@m.xtype_handler(None)
class StringValueAttribute(AbstractRequirementsAttribute):
    """A string value attribute."""

    value = m.StringPOD("value")


@m.xtype_handler(None)
@m.attr_equal("long_name")
class EnumValue(ReqIFElement):
    """An enumeration value for :class:`.EnumerationDataTypeDefinition`."""

    _xmltag = "specifiedValues"

    def __str__(self) -> str:
        return self.long_name


@m.xtype_handler(None)
class EnumerationDataTypeDefinition(ReqIFElement):
    """An enumeration data type definition for requirement types."""

    _xmltag = "ownedDefinitionTypes"

    values = m.DirectProxyAccessor(
        EnumValue, aslist=m.ElementList, single_attr="long_name"
    )


@m.xtype_handler(None)
class AttributeDefinitionEnumeration(ReqIFElement):
    """An enumeration attribute definition for requirement types."""

    _xmltag = "ownedAttributes"

    data_type = m.AttrProxyAccessor(
        EnumerationDataTypeDefinition, "definitionType"
    )
    multi_valued = m.BoolPOD("multiValued")
    """Whether to allow setting multiple values on this attribute."""


@m.xtype_handler(None)
class EnumerationValueAttribute(AbstractRequirementsAttribute):
    """An enumeration attribute."""

    definition = m.AttrProxyAccessor(
        AttributeDefinitionEnumeration,  # type: ignore[arg-type]
        "definition",
    )
    values = m.AttrProxyAccessor(EnumValue, "values", aslist=m.ElementList)

    @property
    def value(self):
        vals = self.values
        if len(vals) > 1:
            raise TypeError("Multi-value enumeration, use `.values` instead")
        if len(vals) == 1:
            return vals[0]
        return None  # TODO return a default value?

    def _repr_value(self) -> str:
        return repr([i.long_name for i in self.values])


@m.attr_equal("long_name")
class AbstractType(ReqIFElement):
    owner = m.ParentAccessor(m.ModelElement)
    attribute_definitions = m.DirectProxyAccessor(
        m.ModelElement,
        (AttributeDefinition, AttributeDefinitionEnumeration),
        aslist=m.MixedElementList,
    )


@m.xtype_handler(None)
class ModuleType(AbstractType):
    """A requirement-module type."""

    _xmltag = "ownedTypes"


@m.xtype_handler(None)
class RelationType(AbstractType):
    """A requirement-relation type."""

    _xmltag = "ownedTypes"


@m.xtype_handler(None)
class RequirementType(AbstractType):
    """A requirement type."""

    _xmltag = "ownedTypes"


@m.xtype_handler(None)
class Requirement(ReqIFElement):
    """A ReqIF Requirement."""

    _xmltag = "ownedRequirements"

    owner = m.ParentAccessor(m.ModelElement)

    chapter_name = m.StringPOD("ReqIFChapterName")
    foreign_id = m.IntPOD("ReqIFForeignID")
    text = m.HTMLStringPOD("ReqIFText")
    attributes = AttributeAccessor()
    type = m.AttrProxyAccessor(RequirementType, "requirementType")

    relations: m.Accessor[AbstractRequirementsRelation]
    related: m.Accessor[m.ModelElement]


@m.xtype_handler(None)
class Folder(Requirement):
    """A folder that stores Requirements."""

    _xmltag = "ownedRequirements"

    folders: m.Accessor
    requirements = m.DirectProxyAccessor(Requirement, aslist=m.ElementList)


class AbstractRequirementsRelation(ReqIFElement):
    _required_attrs = frozenset({"source", "target"})

    type = m.AttrProxyAccessor(RelationType, "relationType")
    source = m.AttrProxyAccessor(Requirement, "source")
    target = m.AttrProxyAccessor(m.ModelElement, "target")

    def _short_repr_(self) -> str:
        direction = ""
        if self.source is not None:
            direction += f" from {self.source._short_repr_()}"
        if self.target is not None:
            direction += f" to {self.target._short_repr_()}"
        return (
            f"<{type(self).__name__} {self.long_name!r}{direction} "
            f"({self.uuid})>"
        )


@m.xtype_handler(None)
class InternalRelation(AbstractRequirementsRelation):
    """A Relation between two requirements."""

    _xmltag = "ownedRelations"


_attr_type_hints = {
    "int": IntegerValueAttribute,
    "integer": IntegerValueAttribute,
    "integervalueattribute": IntegerValueAttribute,
    "str": StringValueAttribute,
    "string": StringValueAttribute,
    "stringvalueattribute": StringValueAttribute,
    "float": RealValueAttribute,
    "real": RealValueAttribute,
    "realvalueattribute": RealValueAttribute,
    "date": DateValueAttribute,
    "datetime": DateValueAttribute,
    "datevalueattribute": DateValueAttribute,
    "bool": BooleanValueAttribute,
    "boolean": BooleanValueAttribute,
    "booleanvalueattribute": BooleanValueAttribute,
    "enum": EnumerationValueAttribute,
    "enumeration": EnumerationValueAttribute,
    "enumerationvalueattribute": EnumerationValueAttribute,
}



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Abstract classes acting as templates for concrete classes.

These base classes are used between different layers.
"""

from capellambse import model as m


class TraceableElement(m.ModelElement):
    """A template for traceable ModelObjects."""

    source = m.AttrProxyAccessor(m.ModelElement, attr="sourceElement")
    target = m.AttrProxyAccessor(m.ModelElement, attr="targetElement")



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from . import capellacommon as capellacommon
from . import capellacore as capellacore
from . import capellamodeller as capellamodeller
from . import cs as cs
from . import fa as fa
from . import interaction as interaction
from . import la as la
from . import modellingcore as modellingcore
from . import oa as oa
from . import pa as pa
from . import sa as sa

import capellambse.model as m  # isort: skip

m.set_accessor(
    capellacommon.State,
    "functions",
    m.ReferenceSearchingAccessor(
        (
            oa.OperationalActivity,
            sa.SystemFunction,
            la.LogicalFunction,
            pa.PhysicalFunction,
        ),
        "available_in_states",
        aslist=m.ElementList,
    ),
)
m.set_accessor(
    m.ModelElement,
    "property_value_packages",
    m.DirectProxyAccessor(
        capellacore.PropertyValuePkg,
        aslist=m.ElementList,
        mapkey="name",
    ),
)

del m



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import capellambse.model as m

from . import la, oa, pa, sa


@m.xtype_handler(None)
class SystemEngineering(m.ModelElement):
    """A system engineering element.

    System engineering is an interdisciplinary approach encompassing the entire
    technical effort to evolve and verify an integrated and life-cycle balanced
    set of system people, product, and process solutions that satisfy customer
    needs.

    Systems engineering encompasses:

    - the technical efforts related to the development, manufacturing,
      verification, deployment, operations, support, disposal of, and user
      training for, systems products and processes;
    - the definition and management of the system configuration;
    - the translation of the system definition into work breakdown structures;
    - and development of information for management decision making.

    [source:MIL-STD 499B standard]
    """

    architectures = m.RoleTagAccessor(
        "ownedArchitectures", m.ModelElement, aslist=m.ElementList
    )

    @property
    def oa(self) -> oa.OperationalAnalysis:
        try:
            return next(
                i
                for i in self.architectures
                if isinstance(i, oa.OperationalAnalysis)
            )
        except StopIteration:
            raise AttributeError(
                f"OperationalAnalysis not found on {self._short_repr_()}"
            ) from None

    @property
    def sa(self) -> sa.SystemAnalysis:
        try:
            return next(
                i
                for i in self.architectures
                if isinstance(i, sa.SystemAnalysis)
            )
        except StopIteration:
            raise AttributeError(
                f"SystemAnalysis not found on {self._short_repr_()}"
            ) from None

    @property
    def la(self) -> la.LogicalArchitecture:
        try:
            return next(
                i
                for i in self.architectures
                if isinstance(i, la.LogicalArchitecture)
            )
        except StopIteration:
            raise AttributeError(
                f"LogicalArchitecture not found on {self._short_repr_()}"
            ) from None

    @property
    def pa(self) -> pa.PhysicalArchitecture:
        try:
            return next(
                i
                for i in self.architectures
                if isinstance(i, pa.PhysicalArchitecture)
            )
        except StopIteration:
            raise AttributeError(
                f"PhysicalArchitecture not found on {self._short_repr_()}"
            ) from None


@m.xtype_handler(None)
class Project(m.ModelElement):
    model_roots = m.RoleTagAccessor(
        "ownedModelRoots", SystemEngineering, aslist=m.ElementList
    )

    @property
    def model_root(self) -> SystemEngineering:
        if self.model_roots:
            return self.model_roots[0]
        return self.model_roots.create()


@m.xtype_handler(None)
class Library(Project):
    """A project that is primarily intended as a library of components."""



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Implementation of objects and relations for Functional Analysis.

Composite Structure objects inheritance tree (taxonomy):

.. diagram:: [CDB] CompositeStructure [Taxonomy]

Composite Structure object-relations map (ontology):

.. diagram:: [CDB] CompositeStructure [Ontology]
"""

import capellambse.model as m

from . import capellacommon, fa, information


@m.xtype_handler(None)
class Part(m.ModelElement):
    """A representation of a physical component."""

    _xmltag = "ownedParts"

    type = m.AttrProxyAccessor(m.ModelElement, "abstractType")

    deployed_parts: m.Accessor


@m.xtype_handler(None)
class ExchangeItemAllocation(m.ModelElement):
    """An allocation of an ExchangeItem to an Interface."""

    item = m.AttrProxyAccessor(information.ExchangeItem, "allocatedItem")


@m.xtype_handler(None)
class Interface(m.ModelElement):
    """An interface."""

    exchange_item_allocations = m.DirectProxyAccessor(
        ExchangeItemAllocation, aslist=m.ElementList
    )


@m.xtype_handler(None)
class InterfacePkg(m.ModelElement):
    """A package that can hold interfaces and exchange items."""

    exchange_items = m.DirectProxyAccessor(
        information.ExchangeItem, aslist=m.ElementList
    )
    interfaces = m.DirectProxyAccessor(Interface, aslist=m.ElementList)

    packages: m.Accessor


@m.xtype_handler(None)
class PhysicalPort(m.ModelElement):
    """A physical port."""

    _xmltag = "ownedFeatures"

    owner = m.ParentAccessor(m.ModelElement)
    links: m.Accessor


@m.xtype_handler(None)
class PhysicalLink(PhysicalPort):
    """A physical link."""

    ends = m.PhysicalLinkEndsAccessor(
        PhysicalPort, "linkEnds", aslist=m.ElementList
    )
    exchanges = m.LinkAccessor[fa.ComponentExchange](
        "ownedComponentExchangeAllocations",
        fa.ComponentExchangeAllocation,
        aslist=m.ElementList,
        attr="targetElement",
        backattr="sourceElement",
    )

    physical_paths: m.Accessor

    source = m.IndexAccessor[PhysicalPort]("ends", 0)
    target = m.IndexAccessor[PhysicalPort]("ends", 1)


@m.xtype_handler(None)
class PhysicalPath(m.ModelElement):
    """A physical path."""

    _xmltag = "ownedPhysicalPath"

    involved_items = m.LinkAccessor[m.ModelElement](
        None,  # FIXME fill in tag
        "org.polarsys.capella.core.data.cs:PhysicalPathInvolvement",
        aslist=m.MixedElementList,
        attr="involved",
    )
    exchanges = m.LinkAccessor[fa.ComponentExchange](
        None,  # FIXME fill in tag
        "org.polarsys.capella.core.data.fa:ComponentExchangeAllocation",
        aslist=m.ElementList,
        attr="targetElement",
    )

    @property
    def involved_links(self) -> m.ElementList[PhysicalLink]:
        return self.involved_items.by_type("PhysicalLink")


class Component(m.ModelElement):
    """A template class for components."""

    is_abstract = m.BoolPOD("abstract")
    """Boolean flag for an abstract Component."""
    is_human = m.BoolPOD("human")
    """Boolean flag for a human Component."""
    is_actor = m.BoolPOD("actor")
    """Boolean flag for an actor Component."""

    owner = m.ParentAccessor(m.ModelElement)
    state_machines = m.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=m.ElementList
    )
    ports = m.DirectProxyAccessor(fa.ComponentPort, aslist=m.ElementList)
    physical_ports = m.DirectProxyAccessor(PhysicalPort, aslist=m.ElementList)
    parts = m.ReferenceSearchingAccessor(Part, "type", aslist=m.ElementList)
    physical_paths = m.DirectProxyAccessor(PhysicalPath, aslist=m.ElementList)
    physical_links = m.DirectProxyAccessor(PhysicalLink, aslist=m.ElementList)
    exchanges = m.DirectProxyAccessor(
        fa.ComponentExchange, aslist=m.ElementList
    )

    related_exchanges = m.ReferenceSearchingAccessor(
        fa.ComponentExchange,
        "source.owner",
        "target.owner",
        aslist=m.ElementList,
    )

    realized_components = m.LinkAccessor["Component"](
        "ownedComponentRealizations",
        "org.polarsys.capella.core.data.cs:ComponentRealization",
        aslist=m.ElementList,
        attr="targetElement",
    )
    realizing_components = m.ReferenceSearchingAccessor["Component"](
        (), "realized_components", aslist=m.ElementList
    )


@m.xtype_handler(None)
class ComponentRealization(m.ModelElement):
    """A realization that links to a component."""

    _xmltag = "ownedComponentRealizations"


class ComponentArchitecture(m.ModelElement):
    """Formerly known as BaseArchitectureLayer."""

    data_package = m.DirectProxyAccessor(information.DataPkg)
    interface_package = m.DirectProxyAccessor(InterfacePkg)
    component_exchange_categories = m.DirectProxyAccessor(
        fa.ComponentExchangeCategory, aslist=m.ElementList
    )

    all_classes = m.DeepProxyAccessor(information.Class, aslist=m.ElementList)
    all_collections = m.DeepProxyAccessor(
        information.Collection, aslist=m.ElementList
    )
    all_unions = m.DeepProxyAccessor(information.Union, aslist=m.ElementList)
    all_enumerations = m.DeepProxyAccessor(
        information.datatype.Enumeration, aslist=m.ElementList
    )
    all_complex_values = m.DeepProxyAccessor(
        information.datavalue.ComplexValue, aslist=m.ElementList
    )
    all_interfaces = m.DeepProxyAccessor(Interface, aslist=m.ElementList)


m.set_accessor(
    InterfacePkg,
    "packages",
    m.DirectProxyAccessor(InterfacePkg, aslist=m.ElementList),
)
m.set_accessor(
    Part,
    "deployed_parts",
    m.LinkAccessor(
        "ownedDeploymentLinks",
        "org.polarsys.capella.core.data.pa.deployment:PartDeploymentLink",
        aslist=m.ElementList,
        attr="deployedElement",
        backattr="location",
    ),
)
m.set_accessor(
    PhysicalPort,
    "links",
    m.ReferenceSearchingAccessor(PhysicalLink, "ends", aslist=m.ElementList),
)
m.set_accessor(
    PhysicalLink,
    "physical_paths",
    m.ReferenceSearchingAccessor(
        PhysicalPath, "involved_items", aslist=m.ElementList
    ),
)
m.set_accessor(
    fa.ComponentExchange,
    "allocating_physical_link",
    m.ReferenceSearchingAccessor(PhysicalLink, "exchanges"),
)
m.set_accessor(
    fa.ComponentExchange,
    "allocating_physical_paths",
    m.ReferenceSearchingAccessor(
        PhysicalPath, "exchanges", aslist=m.ElementList
    ),
)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
import capellambse.model as m

from . import capellacore


class FunctionalChainAbstractCapabilityInvolvement(m.ModelElement): ...


class AbstractCapabilityRealization(m.ModelElement): ...


@m.xtype_handler(None)
class Execution(m.ModelElement):
    """An execution."""

    start = m.AttrProxyAccessor(m.ModelElement, "start")
    finish = m.AttrProxyAccessor(m.ModelElement, "finish")


@m.xtype_handler(None)
class StateFragment(Execution):
    """A state fragment."""

    function = m.AttrProxyAccessor(m.ModelElement, "relatedAbstractFunction")


@m.xtype_handler(None)
class CombinedFragment(Execution):
    """A combined fragment."""

    operator = m.StringPOD("operator")
    operands = m.AttrProxyAccessor(
        m.ModelElement, "referencedOperands", aslist=m.ElementList
    )


@m.xtype_handler(None)
class InstanceRole(m.ModelElement):
    """An instance role."""

    instance = m.AttrProxyAccessor[m.ModelElement](None, "representedInstance")


@m.xtype_handler(None)
class SequenceMessage(m.ModelElement):
    """A sequence message."""

    source = m.AttrProxyAccessor(m.ModelElement, "sendingEnd")
    target = m.AttrProxyAccessor(m.ModelElement, "receivingEnd")


class Event(m.ModelElement):
    """Abstract super class of all events in a Scenario."""


class EventOperation(Event):
    """Abstract super class for events about operations."""

    operation = m.AttrProxyAccessor(m.ModelElement, "operation")


@m.xtype_handler(None)
class ExecutionEvent(Event):
    """An execution event."""


@m.xtype_handler(None)
class EventSentOperation(EventOperation):
    """An event-sent operation."""


@m.xtype_handler(None)
class EventReceiptOperation(EventOperation):
    """An event-receipt operation."""


@m.xtype_handler(None)
class Scenario(m.ModelElement):
    """A scenario that holds instance roles."""

    instance_roles = m.DirectProxyAccessor[m.ModelElement](
        InstanceRole, aslist=m.ElementList
    )
    messages = m.DirectProxyAccessor(SequenceMessage, aslist=m.ElementList)
    events = m.RoleTagAccessor("ownedEvents", aslist=m.MixedElementList)
    fragments = m.RoleTagAccessor(
        "ownedInteractionFragments", aslist=m.MixedElementList
    )
    time_lapses = m.RoleTagAccessor(
        "ownedTimeLapses", aslist=m.MixedElementList
    )
    postcondition = m.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = m.AttrProxyAccessor(capellacore.Constraint, "preCondition")


class InteractionFragment(m.ModelElement):
    """Abstract super class of all interaction fragments in a Scenario."""

    covered = m.AttrProxyAccessor[m.ModelElement](
        None, "coveredInstanceRoles", aslist=m.MixedElementList
    )


@m.xtype_handler(None)
class ExecutionEnd(InteractionFragment):
    """An end for an execution."""

    event = m.AttrProxyAccessor[Event](None, "event")


@m.xtype_handler(None)
class FragmentEnd(InteractionFragment):
    """An end for a fragment."""


@m.xtype_handler(None)
class InteractionOperand(InteractionFragment):
    """An interaction-operand."""

    guard = m.AttrProxyAccessor(capellacore.Constraint, "guard")


@m.xtype_handler(None)
class InteractionState(InteractionFragment):
    """An interaction-state."""

    state = m.AttrProxyAccessor(m.ModelElement, "relatedAbstractState")
    function = m.AttrProxyAccessor(m.ModelElement, "relatedAbstractFunction")


@m.xtype_handler(None)
class MessageEnd(InteractionFragment):
    """A message-end."""

    event = m.AttrProxyAccessor[Event](None, "event")


class Exchange(m.ModelElement):
    """An abstract Exchange."""

    source = m.ParentAccessor(m.ModelElement)


@m.xtype_handler(None)
class AbstractCapabilityExtend(Exchange):
    """An AbstractCapabilityExtend."""

    _xmltag = "extends"

    source = m.ParentAccessor(m.ModelElement)
    target = m.AttrProxyAccessor(m.ModelElement, "extended")


@m.xtype_handler(None)
class AbstractCapabilityInclude(Exchange):
    """An AbstractCapabilityInclude."""

    _xmltag = "includes"

    source = m.ParentAccessor(m.ModelElement)
    target = m.AttrProxyAccessor(m.ModelElement, "included")


@m.xtype_handler(None)
class AbstractCapabilityGeneralization(Exchange):
    """An AbstractCapabilityGeneralization."""

    _xmltag = "superGeneralizations"

    source = m.ParentAccessor(m.ModelElement)
    target = m.AttrProxyAccessor(m.ModelElement, "super")


class AbstractInvolvement(m.ModelElement):
    """An abstract Involvement."""

    source = m.ParentAccessor(m.ModelElement)
    target = m.AttrProxyAccessor(m.ModelElement, "involved")

    involved = m.AttrProxyAccessor(m.ModelElement, "involved")

    @property
    def name(self) -> str:  # type: ignore[override]
        """Return the name."""
        direction = ""
        if self.involved is not None:
            direction = f" to {self.involved.name} ({self.involved.uuid})"

        return f"[{self.__class__.__name__}]{direction}"


@m.xtype_handler(None)
class AbstractFunctionAbstractCapabilityInvolvement(AbstractInvolvement):
    """An abstract CapabilityInvolvement linking to SystemFunctions."""



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = [
    "ModelObject",
    "ModelElement",
    "ElementList",
    "CachedElementList",
    "MixedElementList",
    "ElementListMapKeyView",
    "ElementListMapItemsView",
    "ElementListCouplingMixin",
]

import collections.abc as cabc
import contextlib
import enum
import inspect
import logging
import operator
import re
import textwrap
import typing as t

import markupsafe
import typing_extensions as te
from lxml import etree

import capellambse
from capellambse import helpers

from . import T, U, _descriptors, _pods, _styleclass, _xtype

LOGGER = logging.getLogger(__name__)


_NOT_SPECIFIED = object()
"Used to detect unspecified optional arguments"

_MapFunction: te.TypeAlias = (
    "cabc.Callable[[T], ModelElement | cabc.Iterable[ModelElement]]"
)

_TERMCELL: tuple[int, int] | None = None
_ICON_CACHE: dict[tuple[str, str, int], t.Any] = {}


class ModelObject(t.Protocol):
    """A class that wraps a specific model object.

    Most of the time, you'll want to subclass the concrete
    ``ModelElement`` class. However, some special classes (e.g. AIRD
    diagrams) provide a compatible interface, but it doesn't make sense
    to wrap a specific XML element. This protocol class is used in type
    annotations to catch both "normal" ModelElement subclasses and the
    mentioned special cases.
    """

    @property
    def _model(self) -> capellambse.MelodyModel: ...

    @property
    def _element(self) -> etree._Element: ...

    def __init__(
        self,
        model: capellambse.MelodyModel,
        parent: etree._Element,
        xmltag: str | None,
        /,
        **kw: t.Any,
    ) -> None:
        """Create a new model object.

        Parameters
        ----------
        model
            The model instance.
        parent
            The parent XML element below which to create a new object.
        xmltag
            Override the XML tag to use for this element.
        kw
            Any additional arguments will be used to populate the
            instance attributes. Note that some attributes may be
            required by specific element types at construction time
            (commonly e.g. ``uuid``).
        """

    @classmethod
    def from_model(
        cls, model: capellambse.MelodyModel, element: t.Any
    ) -> ModelObject:
        """Instantiate a ModelObject from existing model elements."""


class ModelElement:
    """A model element.

    This is the common base class for all elements of a model. In terms
    of the metamodel, it combines the role of
    ``modellingcore.ModelElement`` and all its superclasses; references
    to any superclass should be modified to use this class instead.
    """

    uuid = _pods.StringPOD("id", writable=False)
    """The universally unique identifier of this object.

    This attribute is automatically populated when the object is
    instantiated, and cannot be changed afterwards. It is however
    possible to specify the UUID when instantiating the object, in which
    case it will be used instead of generating a new one.

    The UUID may be used in hashmaps and the like, as it is guaranteed
    to not change during the object's lifetime and across model
    save/reload cycles.
    """

    sid = _pods.StringPOD("sid")
    """The unique system identifier of this object."""

    xtype = property(lambda self: helpers.xtype_of(self._element))
    name = _pods.StringPOD("name")
    description = _pods.HTMLStringPOD("description")
    summary = _pods.StringPOD("summary")
    diagrams: _descriptors.Accessor[capellambse.model.diagram.Diagram]
    diagrams = property(  # type: ignore[assignment]
        lambda self: self._model.diagrams.by_target(self)
    )
    visible_on_diagrams = property(
        lambda self: self._model.diagrams.by_semantic_nodes(self)
    )

    parent: _descriptors.ParentAccessor
    constraints: _descriptors.Accessor
    property_value_packages: _descriptors.Accessor

    _required_attrs = frozenset({"uuid", "xtype"})
    _xmltag: str | None = None

    __hash__ = None  # type: ignore[assignment]
    """Disable hashing by default on the base class.

    The ``__hash__`` contract states that two objects that compare equal
    must have the same hash value, and that the hash value must not
    change over an object's lifetime.

    Some subclasses of ``ModelElement`` which encapsulate a piece of
    plain old data compare the same as one of their attributes, which
    means they would have to also hash to the same value as that
    attribute.

    However, since ModelElement instances are mutable, that attribute
    could change at any time, which would result in those instances
    ending up in the wrong hash bucket, thus breaking lookups.

    For this reason, and to avoid inconsistent behavior where some
    classes are hashable and some are not, hashing of ModelObjects is
    generally disabled.
    """

    @property
    def progress_status(self) -> str:
        uuid = self._element.get("status")
        if uuid is None:
            return "NOT_SET"

        return self.from_model(self._model, self._model._loader[uuid]).name

    @classmethod
    def from_model(
        cls, model: capellambse.MelodyModel, element: etree._Element
    ) -> te.Self:
        """Wrap an existing model object.

        Parameters
        ----------
        model
            The MelodyModel instance
        element
            The XML element to wrap

        Returns
        -------
        ModelElement
            An instance of ModelElement (or a more appropriate subclass,
            if any) that wraps the given XML element.
        """
        class_ = cls
        if class_ is ModelElement:
            xtype = helpers.xtype_of(element)
            if xtype is not None:
                with contextlib.suppress(KeyError):
                    class_ = _xtype.XTYPE_HANDLERS[None][xtype]
            if class_ is not cls:
                return class_.from_model(model, element)
        self = class_.__new__(class_)
        self._model = model
        self._element = element
        return self

    @property
    def layer(self) -> capellambse.metamodel.cs.ComponentArchitecture:
        """Find the layer that this element belongs to.

        Note that an architectural layer normally does not itself have a
        parent layer.

        Raises
        ------
        AttributeError
            Raised if this element is not nested below a layer.
        """
        import capellambse.metamodel as mm

        obj: ModelElement | None = self
        assert obj is not None
        while obj := getattr(obj, "parent", None):
            if isinstance(obj, mm.cs.ComponentArchitecture):
                return obj
        raise AttributeError(
            f"No parent layer found for {self._short_repr_()}"
        )

    def __init__(
        self,
        model: capellambse.MelodyModel,
        parent: etree._Element,
        xmltag: str | None = None,
        /,
        *,
        uuid: str,
        **kw: t.Any,
    ) -> None:
        all_required_attrs: set[str] = set()
        for basecls in type(self).mro():
            all_required_attrs |= getattr(
                basecls, "_required_attrs", frozenset()
            )
        missing_attrs = all_required_attrs - frozenset(kw) - {"uuid"}
        if missing_attrs:
            mattrs = ", ".join(sorted(missing_attrs))
            raise TypeError(f"Missing required keyword arguments: {mattrs}")

        super().__init__()
        if xmltag is None:
            xmltag = self._xmltag
        if xmltag is None:
            raise TypeError(
                f"Cannot instantiate {type(self).__name__} directly"
            )
        self._model = model
        self._element: etree._Element = etree.Element(xmltag)
        parent.append(self._element)
        try:
            self.uuid = uuid
            for key, val in kw.items():
                if key == "xtype":
                    self._element.set(helpers.ATT_XT, val)
                elif not isinstance(
                    getattr(type(self), key),
                    _descriptors.Accessor | _pods.BasePOD,
                ):
                    raise TypeError(
                        f"Cannot set {key!r} on {type(self).__name__}"
                    )
                else:
                    setattr(self, key, val)
            self._model._loader.idcache_index(self._element)
        except BaseException:
            parent.remove(self._element)
            raise

    def __setattr__(self, attr: str, value: t.Any) -> None:
        if attr.startswith("_") or hasattr(type(self), attr):
            super().__setattr__(attr, value)
        else:
            raise AttributeError(
                f"{attr!r} isn't defined on {type(self).__name__}"
            )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._element is other._element

    def __dir__(self) -> list[str]:
        badacc = (_descriptors.DeprecatedAccessor,)
        cls = type(self)
        attrs: list[str] = []
        for i in super().__dir__():
            try:
                acc = getattr(cls, i)
            except Exception:
                continue
            if isinstance(acc, badacc):
                continue
            if isinstance(acc, _descriptors.Alias) and acc.dirhide:
                continue
            try:
                if getattr(acc, "__deprecated__", None):
                    continue
            except Exception:
                continue
            attrs.append(i)
        return attrs

    def __repr__(self) -> str:  # pragma: no cover
        header = self._short_repr_()

        attrs: list[str] = []
        for attr in dir(self):
            if attr.startswith("_"):
                continue

            acc = getattr(type(self), attr, None)
            if isinstance(acc, _descriptors.ReferenceSearchingAccessor):
                classes = ", ".join(i.__name__ for i in acc.target_classes)
                attrs.append(
                    f".{attr} = ... # backreference to {classes}"
                    " - omitted: can be slow to compute"
                )
                continue

            if attr.startswith("all_"):
                attrs.append(f".{attr} = ... # omitted")
                continue

            try:
                value = getattr(self, attr)
            except Exception:
                continue

            if inspect.ismethod(value):
                continue

            if hasattr(value, "_short_repr_"):
                value_repr = f"{value._short_repr_()}"
            else:
                value_repr = textwrap.shorten(repr(value), 250)

            prefix = f".{attr} = "
            blankprefix = " " * len(prefix)
            value_repr = "\n".join(
                (prefix, blankprefix)[bool(i)] + line
                for i, line in enumerate(value_repr.splitlines() or [""])
            )
            attrs.append(value_repr)

        attr_text = "\n".join(attrs)
        return f"{header}\n{attr_text}"

    def _short_repr_(self) -> str:
        if type(self) is ModelElement:
            mytype = f"Model element ({self.xtype})"
        else:
            mytype = type(self).__name__

        if self.name:
            name = f" {self.name!r}"
        else:
            name = ""

        if capellambse.model.diagram.REPR_DRAW:
            global _TERMCELL
            if _TERMCELL is None:
                try:
                    _TERMCELL = helpers.get_term_cell_size()
                except ValueError as err:
                    LOGGER.warning("Cannot determine term cell size: %s", err)
                    _TERMCELL = (0, 0)
            size = (_TERMCELL[1] or 13) - 2
            icon = self._get_icon("termgraphics", size=size) or b""
            assert isinstance(icon, bytes)
        else:
            icon = b""
        return f"<{icon.decode()}{mytype}{name} ({self.uuid})>"

    def __html__(self) -> markupsafe.Markup:
        fragments: list[str] = []
        escape = markupsafe.Markup.escape

        try:
            icon = self._get_icon("datauri_svg", size=20)
        except Exception:
            icon = None

        fragments.append("<h1>")
        if icon:
            fragments.append(
                f'<img src="{icon}" alt="" width="20" height="20"> '
            )
        if type(self) is ModelElement:
            fragments.append("Model element")
        else:
            fragments.append(escape(self.name or type(self).__name__))
        fragments.append(' <span style="font-size: 70%;">(')
        fragments.append(escape(self.xtype))
        fragments.append(")</span></h1>")

        fragments.append("<table>")
        for attr in dir(self):
            if attr.startswith("_"):
                continue

            acc = getattr(type(self), attr, None)
            if isinstance(acc, _descriptors.ReferenceSearchingAccessor):
                classes = ", ".join(i.__name__ for i in acc.target_classes)
                fragments.append('<tr><th style="text-align: right;">')
                fragments.append(escape(attr))
                fragments.append('</th><td style="text-align: left;"><em>')
                fragments.append(f"Backreference to {escape(classes)}")
                fragments.append(" - omitted: can be slow to compute.")
                fragments.append(" Display this property directly to show.")
                fragments.append("</em></td></tr>")
                continue

            try:
                value = getattr(self, attr)
            except Exception:
                continue

            if inspect.ismethod(value):
                continue

            fragments.append('<tr><th style="text-align: right;">')
            fragments.append(escape(attr))
            fragments.append('</th><td style="text-align: left;">')

            if hasattr(value, "_short_html_"):
                fragments.append(value._short_html_())
            elif isinstance(value, str):
                fragments.append(escape(value))
            else:
                value = repr(value)
                if len(value) > 250:
                    value = value[:250] + " [...]"
                fragments.append("<em>")
                fragments.append(escape(value))
                fragments.append("</em>")
            fragments.append("</td></tr>")
        fragments.append("</table>")
        return markupsafe.Markup("".join(fragments))

    def _short_html_(self) -> markupsafe.Markup:
        try:
            icon = self._get_icon("datauri_svg", size=15) or ""
        except Exception:
            icon = ""
        else:
            assert isinstance(icon, str)
        value = getattr(self, "value", "")
        if hasattr(value, "_short_html_"):
            value = value._short_html_()
        return helpers.make_short_html(
            type(self).__name__,
            self.uuid,
            self.name,
            value,
            icon=icon,
            iconsize=15,
        )

    def _repr_html_(self) -> str:
        return self.__html__()

    def _get_styleclass(self) -> str:
        """Return the styleclass of this object.

        :meta public:

        The styleclass determines which set of styles gets applied when
        drawing this object in a diagram.
        """
        return _styleclass.get_styleclass(self)

    def _get_icon(self, format: str, /, *, size: int = 16) -> t.Any | None:
        """Render a small icon for this object.

        :meta public:

        This is the same icon that is also used in diagrams.

        Parameters
        ----------
        format
            The format to use.

            This uses the same format conversion machinery as diagrams, but
            starts with the *svg* format. This means that *svg* and every
            format directly or indirectly derived from it are supported,
            including *png*, *datauri_svg* and others.
        size
            Return the icon scaled to this horizontal and vertical size
            in pixels. This may yield higher quality results compared to
            scaling the returned icon, especially when using raster
            image formats.

        Returns
        -------
        Any | None
            The icon (usually as str or bytes object), or None if no
            icon could be found.
        """
        from capellambse.diagram import get_icon

        sc = self._get_styleclass()
        try:
            return _ICON_CACHE[sc, format, size]
        except KeyError:
            pass

        try:
            data: t.Any = get_icon(sc, size=size)
        except ValueError:
            return None

        if format != "svg":
            data = capellambse.model.diagram.convert_format(
                "svg", format, data
            )
        _ICON_CACHE[sc, format, size] = data
        return data

    if t.TYPE_CHECKING:

        def __getattr__(self, attr: str) -> t.Any:
            """Account for extension attributes in static type checks."""


class ElementList(cabc.MutableSequence[T], t.Generic[T]):
    """Provides access to elements without affecting the underlying model."""

    __slots__ = (
        "_elemclass",
        "_elements",
        "_ElementList__mapkey",
        "_ElementList__mapvalue",
        "_model",
    )

    def is_coupled(self) -> bool:
        return False

    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: type[T] | None = None,
        *,
        mapkey: str | None = None,
        mapvalue: str | None = None,
    ) -> None:
        assert None not in elements
        self._model = model
        self._elements = elements
        if elemclass is not None:
            self._elemclass = elemclass
        else:
            self._elemclass = ModelElement  # type: ignore[assignment]

        if not mapkey:
            self.__mapkey: str | None = None
            self.__mapvalue: str | None = None
        else:
            self.__mapkey = mapkey
            self.__mapvalue = mapvalue

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, cabc.Sequence):
            return NotImplemented
        return len(self) == len(other) and all(
            ours == theirs for ours, theirs in zip(self, other, strict=True)
        )

    def __add(
        self, other: object, *, reflected: bool = False
    ) -> ElementList[T]:
        if not isinstance(other, ElementList):
            return NotImplemented
        if self._model is not other._model:
            raise ValueError("Cannot add ElementLists from different models")

        if self._elemclass is other._elemclass is not ModelElement:
            listclass: type[ElementList] = ElementList
            elemclass: type[ModelObject] = self._elemclass
        else:
            listclass = MixedElementList
            elemclass = ModelElement

        if not reflected:
            elements = self._elements + other._elements
        else:
            elements = other._elements + self._elements

        return listclass(self._model, elements, elemclass)

    def __add__(self, other: object) -> ElementList[T]:
        return self.__add(other)

    def __radd__(self, other: object) -> ElementList[T]:
        return self.__add(other, reflected=True)

    def __sub(
        self, other: object, *, reflected: bool = False
    ) -> ElementList[T]:
        if not isinstance(other, cabc.Sequence):
            return NotImplemented

        if reflected:
            if isinstance(other, ElementList):
                objclass = other._elemclass
            else:
                objclass = ModelElement
        else:
            objclass = self._elemclass

        base: cabc.Sequence[t.Any]
        if not reflected:
            base = self
            excluded = {getattr(i, "uuid", None) for i in other}
        else:
            base = other
            excluded = {getattr(i, "uuid", None) for i in self}

        return ElementList(
            self._model,
            [i._element for i in base if i.uuid not in excluded],
            objclass,
        )

    def __sub__(self, other: object) -> ElementList[T]:
        """Return a new list without elements found in ``other``."""
        return self.__sub(other)

    def __rsub__(self, other: object) -> ElementList[T]:
        return self.__sub(other, reflected=True)

    def __len__(self) -> int:
        return len(self._elements)

    @t.overload
    def __getitem__(self, idx: int) -> T: ...
    @t.overload
    def __getitem__(self, idx: slice) -> ElementList[T]: ...
    @t.overload
    def __getitem__(self, idx: str) -> t.Any: ...
    def __getitem__(self, idx: int | slice | str) -> t.Any:
        if isinstance(idx, slice):
            return self._newlist(self._elements[idx])
        if isinstance(idx, str):
            obj = self._map_find(idx)
            return self._map_getvalue(obj)
        return self._elemclass.from_model(self._model, self._elements[idx])

    @t.overload
    def __setitem__(self, index: int, value: T) -> None: ...
    @t.overload
    def __setitem__(self, index: slice, value: cabc.Iterable[T]) -> None: ...
    @t.overload
    def __setitem__(self, index: str, value: t.Any) -> None: ...
    def __setitem__(self, index: int | slice | str, value: t.Any) -> None:
        if isinstance(index, slice):
            del self[index]
            for i, element in enumerate(value, start=index.start):
                self.insert(i, element)
        elif isinstance(index, int):
            del self[index]
            self.insert(index, value)
        else:
            obj = self._map_find(index)
            self._map_setvalue(obj, value)

    def __delitem__(self, index: int | slice) -> None:
        del self._elements[index]

    def __contains__(self, obj: t.Any) -> bool:
        elem = getattr(obj, "_element", None)
        if isinstance(elem, etree._Element):
            return obj._element in self._elements
        return any(i == obj for i in self)

    def __getattr__(self, attr: str) -> _ListFilter:
        if attr.startswith("by_"):
            attr = attr[len("by_") :]
            if attr in {"name", "uuid"}:
                return _ListFilter(self, attr, single=True)
            return _ListFilter(self, attr)

        if attr.startswith("exclude_") and attr.endswith("s"):
            attr = attr[len("exclude_") : -len("s")]
            return _ListFilter(self, attr, positive=False)

        return getattr(super(), attr)

    def __dir__(self) -> list[str]:  # pragma: no cover
        no_dir_attr = re.compile(r"^(_|as_|pvmt$|diagrams?$)")

        def filterable_attrs() -> cabc.Iterator[str]:
            for obj in self:
                try:
                    obj_attrs = dir(obj)
                except Exception:
                    continue
                for attr in obj_attrs:
                    if not no_dir_attr.search(attr) and isinstance(
                        getattr(obj, attr), str
                    ):
                        yield f"by_{attr}"
                        yield f"exclude_{attr}s"

        attrs = list(super().__dir__())
        attrs.extend(filterable_attrs())
        return attrs

    def __repr__(self) -> str:  # pragma: no cover
        if not self:
            return "[]"

        items: list[str] = []
        for i, item in enumerate(self):
            if hasattr(item, "_short_repr_"):
                item_repr = item._short_repr_()
            else:
                item_repr = repr(item)
            repr_line = item_repr.splitlines() or [""]
            prefix = f"[{i}] "
            repr_line[0] = prefix + repr_line[0]
            prefix = " " * len(prefix)
            repr_line[1:] = [prefix + i for i in repr_line[1:]]
            items.append("\n".join(repr_line))
        return "\n".join(items)

    def _short_repr_(self) -> str:
        return repr(self)

    def __html__(self) -> markupsafe.Markup:
        if not self:
            return markupsafe.Markup("<p><em>(Empty list)</em></p>")

        fragments = ['<ol start="0" style="text-align: left;">']
        for i in self:
            assert hasattr(i, "_short_html_")
            fragments.append(f"<li>{i._short_html_()}</li>")
        fragments.append("</ol>")
        return markupsafe.Markup("".join(fragments))

    def _short_html_(self) -> markupsafe.Markup:
        return self.__html__()

    def _repr_html_(self) -> str:
        return self.__html__()

    def _mapkey(self, obj: T) -> t.Any:
        if self.__mapkey is None:
            raise TypeError("This list cannot act as a mapping")

        mapkey = operator.attrgetter(self.__mapkey)
        try:
            return mapkey(obj)
        except AttributeError:
            return None

    def _map_find(self, key: str) -> T:
        """Find the target object of a mapping operation.

        When this list acts as a mapping (like ``some_list["key"]``),
        this method finds the target object associated with the
        ``"key"``.

        See Also
        --------
        :meth:`_map_getvalue` and :meth:`_map_setvalue`
            Get or set the mapping value behind the target object.
        """
        if self.__mapkey is None:
            raise TypeError("This list cannot act as a mapping")

        mapkey = operator.attrgetter(self.__mapkey)
        candidates = [i for i in self if mapkey(i) == key]
        if len(candidates) > 1:
            raise ValueError(f"Multiple matches for key {key!r}")
        if not candidates:
            raise KeyError(key)
        return candidates[0]

    def _map_getvalue(self, obj: T) -> t.Any:
        """Get the mapping value from the target object."""
        if not self.__mapvalue:
            return obj
        getvalue = operator.attrgetter(self.__mapvalue)
        return getvalue(obj)

    def _map_setvalue(self, obj: T, value: t.Any) -> None:
        """Set a new mapping value on the target object."""
        if not self.__mapvalue:
            self[self.index(obj)] = value
            return

        key = self.__mapvalue.rsplit(".", maxsplit=1)
        if len(key) == 1:
            target: t.Any = obj
        else:
            target = operator.attrgetter(key[0])(obj)

        setattr(target, key[-1], value)

    def _newlist(self, elements: list[etree._Element]) -> ElementList[T]:
        listtype = self._newlist_type()
        return listtype(
            self._model,
            elements,
            self._elemclass,
            mapkey=self.__mapkey,
            mapvalue=self.__mapvalue,
        )

    def _newlist_type(self) -> type[ElementList[T]]:
        return type(self)

    @t.overload
    def get(self, key: str) -> T | None: ...
    @t.overload
    def get(self, key: str, default: U) -> T | U: ...
    def get(self, key: str, default: t.Any = None) -> t.Any:
        try:
            return self[key]
        except KeyError:
            return default

    def insert(self, index: int, value: T) -> None:
        elm: etree._Element = value._element
        self._elements.insert(index, elm)

    def create(self, typehint: str | None = None, /, **kw: t.Any) -> T:
        del typehint, kw
        raise TypeError("Cannot create elements: List is not coupled")

    def create_singleattr(self, arg: t.Any) -> T:
        del arg
        raise TypeError("Cannot create elements: List is not coupled")

    def delete_all(self, **kw: t.Any) -> None:
        """Delete all matching objects from the model."""
        indices: list[int] = []
        for i, obj in enumerate(self):
            if all(getattr(obj, k) == v for k, v in kw.items()):
                indices.append(i)

        for index in reversed(indices):
            del self[index]

    def items(self) -> ElementListMapItemsView[T]:
        return ElementListMapItemsView(self)

    def keys(self) -> ElementListMapKeyView:
        return ElementListMapKeyView(self)

    def values(self) -> ElementList[T]:
        return self

    def filter(
        self, predicate: str | cabc.Callable[[T], bool]
    ) -> ElementList[T]:
        """Filter this list with a custom predicate.

        The predicate may be the name of an attribute or a callable,
        which will be called on each list item. If the attribute value
        or the callable's return value is truthy, the item is included
        in the resulting list.

        When specifying the name of an attribute, nested attributes can
        be chained using ``.``, like ``"parent.name"`` (which would
        pick all elements whose ``parent`` has a non-empty ``name``).
        """
        if isinstance(predicate, str):
            predicate = operator.attrgetter(predicate)
        return self._newlist([i._element for i in self if predicate(i)])

    def map(self, attr: str | _MapFunction[T]) -> ElementList:
        """Apply a function to each element in this list.

        If the argument is a string, it is interpreted as an attribute
        name, and the value of that attribute is returned for each
        element. Nested attribute names can be chained with ``.``.

        If the argument is a callable, it is called for each element,
        and the return value is included in the result. If the callable
        returns a sequence, the sequence is flattened into the result.

        Duplicate values and Nones are always filtered out.

        It is an error if a callable returns something that is not a
        model element or a flat sequence of model elements.
        """
        if isinstance(attr, str):
            if "." in attr:
                mapped: ElementList[t.Any] = self
                for a in attr.split("."):
                    mapped = mapped.map(operator.attrgetter(a))
                return mapped

            attr = operator.attrgetter(attr)
        newelems: list[etree._Element] = []
        newuuids: set[str] = set()
        classes: set[type[ModelElement]] = set()
        for i in self:
            try:
                value = attr(i)
            except AttributeError:
                continue

            if not isinstance(value, cabc.Iterable):
                value = [value]

            for v in value:  # type: ignore[union-attr] # false-positive
                if v is None:
                    continue
                if isinstance(v, ModelElement):
                    if v.uuid in newuuids:
                        continue
                    newuuids.add(v.uuid)
                    newelems.append(v._element)
                    classes.add(type(v))
                else:
                    raise TypeError(
                        f"Map function must return a model element or a list"
                        f" of model elements, not {v!r}"
                    )

        if len(classes) == 1:
            return ElementList(self._model, newelems, classes.pop())
        return MixedElementList(self._model, newelems)


class _ListFilter(t.Generic[T]):
    """Filters this list based on an extractor function."""

    __slots__ = ("_attr", "_parent", "_positive", "_single")

    def __init__(
        self,
        parent: ElementList[T],
        attr: str,
        *,
        positive: bool = True,
        single: bool = False,
    ) -> None:
        """Create a filter object.

        If the extractor returns an :class:`enum.Enum` member for any
        element, the enum member's name will be used for comparisons
        against any ``values``.

        Parameters
        ----------
        parent
            Reference to the :class:`ElementList` this filter should
            operate on
        attr
            The attribute on list members to filter on
        extract_key
            Callable that extracts the key from an element
        positive
            Use elements that match (True) or don't (False)
        single
            When listing all matches, return a single element
            instead. If multiple elements match, it is an error; if
            none match, a ``KeyError`` is raised. Can be overridden
            at call time.
        """
        self._attr = attr
        self._parent = parent
        self._positive = positive
        self._single = single

    def extract_key(self, element: T) -> t.Any:
        extractor = operator.attrgetter(self._attr)
        value = extractor(element)
        if isinstance(value, enum.Enum):
            value = value.name
        return value

    def make_values_container(self, *values: t.Any) -> cabc.Iterable[t.Any]:
        return values

    def ismatch(self, element: T, valueset: cabc.Iterable[t.Any]) -> bool:
        try:
            value = self.extract_key(element)
        except AttributeError:
            return False

        if isinstance(value, str) or not isinstance(value, cabc.Iterable):
            return self._positive == (value in valueset)
        return self._positive == any(v in value for v in valueset)

    @t.overload
    def __call__(self, *values: t.Any, single: t.Literal[True]) -> T: ...
    @t.overload
    def __call__(
        self, *values: t.Any, single: t.Literal[False]
    ) -> ElementList[T]: ...
    @t.overload
    def __call__(
        self, *values: t.Any, single: bool | None = None
    ) -> T | ElementList[T]: ...
    def __call__(
        self, *values: t.Any, single: bool | None = None
    ) -> T | ElementList[T]:
        """List all elements that match this filter.

        Parameters
        ----------
        values
            The values to match against.
        single
            If not ``None``, overrides the ``single`` argument to
            the constructor for this filter call.
        """
        if single is None:
            single = self._single
        valueset = self.make_values_container(*values)
        indices = []
        elements = []
        for i, elm in enumerate(self._parent):
            if self.ismatch(elm, valueset):
                indices.append(i)
                elements.append(self._parent._elements[i])

        if not single:
            return self._parent._newlist(elements)
        if len(elements) > 1:
            value = values[0] if len(values) == 1 else values
            raise KeyError(f"Multiple matches for {value!r}")
        if len(elements) == 0:
            raise KeyError(values[0] if len(values) == 1 else values)
        return self._parent[indices[0]]  # Ensure proper construction

    def __iter__(self) -> cabc.Iterator[t.Any]:
        """Yield values that result in a non-empty list when filtered for.

        The returned iterator yields all values that, when given to
        :meth:`__call__`, will result in a non-empty list being
        returned. Consequently, if the original list was empty, this
        iterator will yield no values.

        The order in which the values are yielded is undefined.
        """
        yielded: set[t.Any] = set()

        for elm in self._parent:
            key = self.extract_key(elm)
            if key not in yielded:
                yield key
                yielded.add(key)

    def __contains__(self, value: t.Any) -> bool:
        valueset = self.make_values_container(value)
        return any(self.ismatch(elm, valueset) for elm in self._parent)

    def __getattr__(self, attr: str) -> te.Self:
        if attr.startswith("_"):
            raise AttributeError(f"Invalid filter attribute name: {attr}")
        return type(self)(
            self._parent,
            f"{self._attr}.{attr}",
            positive=self._positive,
            single=self._single,
        )


class _LowercaseListFilter(_ListFilter[T], t.Generic[T]):
    def extract_key(self, element: T) -> t.Any:
        value = super().extract_key(element)
        assert isinstance(value, str)
        return value.lower()

    def make_values_container(self, *values: t.Any) -> cabc.Iterable[t.Any]:
        return tuple(map(operator.methodcaller("lower"), values))


class CachedElementList(ElementList[T], t.Generic[T]):
    """An ElementList that caches the constructed proxies by UUID."""

    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: type[T],
        *,
        cacheattr: str | None = None,
        **kw: t.Any,
    ) -> None:
        """Create a CachedElementList.

        Parameters
        ----------
        model
            The model that all elements are a part of.
        elements
            The members of this list.
        elemclass
            The :class:`ModelElement` subclass to use for reconstructing
            elements.
        cacheattr
            The attribute on the ``model`` to use as cache.
        **kw
            Additional arguments are passed to the superclass.
        """
        super().__init__(model, elements, elemclass, **kw)
        self.cacheattr = cacheattr

    def __getitem__(self, key):
        elem = super().__getitem__(key)
        if self.cacheattr and not isinstance(elem, ElementList):
            try:
                cache = getattr(self._model, self.cacheattr)
            except AttributeError:
                cache = {}
                setattr(self._model, self.cacheattr, cache)
            elem = cache.setdefault(elem.uuid, elem)
        return elem

    def _newlist(self, elements: list[etree._Element]) -> ElementList[T]:
        newlist = super()._newlist(elements)
        assert isinstance(newlist, CachedElementList)
        newlist.cacheattr = self.cacheattr
        return newlist


class MixedElementList(ElementList[ModelElement]):
    """ElementList that handles proxies using ``XTYPE_HANDLERS``."""

    def __init__(
        self,
        model: capellambse.MelodyModel,
        elements: list[etree._Element],
        elemclass: t.Any = None,
        **kw: t.Any,
    ) -> None:
        """Create a MixedElementList.

        Parameters
        ----------
        model
            The model that all elements are a part of.
        elements
            The members of this list.
        elemclass
            Ignored; provided for drop-in compatibility.
        **kw
            Additional arguments are passed to the superclass.
        """
        del elemclass
        super().__init__(model, elements, ModelElement, **kw)

    def __getattr__(self, attr: str) -> _ListFilter[ModelElement]:
        if attr == "by_type":
            return _LowercaseListFilter(self, "__class__.__name__")
        return super().__getattr__(attr)

    def __dir__(self) -> list[str]:  # pragma: no cover
        return [*super().__dir__(), "by_type", "exclude_types"]


class ElementListMapKeyView(cabc.Sequence):
    def __init__(self, parent, /) -> None:
        self.__parent = parent

    @t.overload
    def __getitem__(self, idx: int) -> t.Any: ...
    @t.overload
    def __getitem__(self, idx: slice) -> list: ...
    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return [self.__parent._mapkey(i) for i in self.__parent[idx]]
        return self.__parent._mapkey(self.__parent[idx])

    def __len__(self) -> int:
        return len(self.__parent)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({list(self)!r})"


class ElementListMapItemsView(cabc.Sequence[tuple[t.Any, T]], t.Generic[T]):
    def __init__(self, parent, /) -> None:
        self.__parent = parent

    @t.overload
    def __getitem__(self, idx: int) -> tuple[t.Any, T]: ...
    @t.overload
    def __getitem__(self, idx: slice) -> list[tuple[t.Any, T]]: ...
    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return [
                (self.__parent._mapkey(i), self.__parent._map_getvalue(i))
                for i in self.__parent[idx]
            ]
        obj = self.__parent[idx]
        return (self.__parent._mapkey(obj), self.__parent._map_getvalue(obj))

    def __len__(self) -> int:
        return len(self.__parent)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({list(self)!r})"


class ElementListCouplingMixin(ElementList[T], t.Generic[T]):
    """Couples an ElementList with an Accessor to enable write support.

    This class is meant to be subclassed further, where the subclass has
    both this class and the originally intended one as base classes (but
    no other ones, i.e. there must be exactly two bases). The Accessor
    then inserts itself as the ``_accessor`` class variable on the new
    subclass. This allows the mixed-in methods to delegate actual model
    modifications to the Accessor.
    """

    _accessor: _descriptors.WritableAccessor[T]

    def is_coupled(self) -> bool:
        return True

    def __init__(
        self,
        *args: t.Any,
        parent: ModelObject,
        fixed_length: int = 0,
        **kw: t.Any,
    ) -> None:
        assert type(self)._accessor

        super().__init__(*args, **kw)
        self._parent = parent
        self.fixed_length = fixed_length

    @t.overload
    def __setitem__(self, index: int, value: T) -> None: ...
    @t.overload
    def __setitem__(self, index: slice, value: cabc.Iterable[T]) -> None: ...
    @t.overload
    def __setitem__(self, index: str, value: t.Any) -> None: ...
    def __setitem__(self, index: int | slice | str, value: t.Any) -> None:
        assert self._parent is not None
        accessor = type(self)._accessor
        assert isinstance(accessor, _descriptors.WritableAccessor)

        if not isinstance(index, int | slice):
            super().__setitem__(index, value)
            return

        new_objs = list(self)
        new_objs[index] = value

        if self.fixed_length and len(new_objs) != self.fixed_length:
            raise TypeError(
                f"Cannot set: List must stay at length {self.fixed_length}"
            )

        accessor.__set__(self._parent, new_objs)

    def __delitem__(self, index: int | slice) -> None:
        if self.fixed_length and len(self) <= self.fixed_length:
            raise TypeError("Cannot delete from a fixed-length list")

        assert self._parent is not None
        accessor = type(self)._accessor
        assert isinstance(accessor, _descriptors.WritableAccessor)
        if not isinstance(index, slice):
            index = slice(index, index + 1 or None)
        for obj in self[index]:
            accessor.delete(self, obj)
        super().__delitem__(index)

    def _newlist_type(self) -> type[ElementList[T]]:
        assert len(type(self).__bases__) == 2
        assert type(self).__bases__[0] is ElementListCouplingMixin
        return type(self).__bases__[1]

    def create(self, typehint: str | None = None, /, **kw: t.Any) -> T:
        """Make a new model object (instance of ModelElement).

        Instead of specifying the full ``xsi:type`` including the
        namespace, you can also pass in just the part after the ``:``
        separator. If this is unambiguous, the appropriate
        layer-specific type will be selected automatically.

        This method can be called with or without the ``layertype``
        argument. If a layertype is not given, all layers will be tried
        to find an appropriate ``xsi:type`` handler. Note that setting
        the layertype to ``None`` explicitly is different from not
        specifying it at all; ``None`` tries only the "Transverse
        modelling" type elements.

        Parameters
        ----------
        typehint
            Hints for finding the correct type of element to create. Can
            either be a full or shortened ``xsi:type`` string, or an
            abbreviation defined by the specific Accessor instance.
        kw
            Initialize the properties of the new object. Depending on
            the object, some attributes may be required.
        """
        if self.fixed_length and len(self) >= self.fixed_length:
            raise TypeError("Cannot create elements in a fixed-length list")

        assert self._parent is not None
        acc = type(self)._accessor
        assert isinstance(acc, _descriptors.WritableAccessor)
        newobj = acc.create(self, typehint, **kw)
        try:
            acc.insert(self, len(self), newobj)
            super().insert(len(self), newobj)
        except:
            self._parent._element.remove(newobj._element)
            raise
        return newobj

    def create_singleattr(self, arg: t.Any) -> T:
        """Make a new model object (instance of ModelElement).

        This new object has only one interesting attribute.

        See Also
        --------
        :meth:`ElementListCouplingMixin.create` :
            More details on how elements are created.
        :meth:`WritableAccessor.create_singleattr` :
            The method to override in Accessors in order to implement
            this operation.
        """
        if self.fixed_length and len(self) >= self.fixed_length:
            raise TypeError("Cannot create elements in a fixed-length list")

        assert self._parent is not None
        acc = type(self)._accessor
        assert isinstance(acc, _descriptors.WritableAccessor)
        newobj = acc.create_singleattr(self, arg)
        try:
            acc.insert(self, len(self), newobj)
            super().insert(len(self), newobj)
        except:
            self._parent._element.remove(newobj._element)
            raise
        return newobj

    def insert(self, index: int, value: T) -> None:
        if self.fixed_length and len(self) >= self.fixed_length:
            raise TypeError("Cannot insert into a fixed-length list")

        assert self._parent is not None
        acc = type(self)._accessor
        assert isinstance(acc, _descriptors.WritableAccessor)
        acc.insert(self, index, value)
        super().insert(index, value)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Tools for the Operational Analysis layer.

.. diagram:: [CDB] OA ORM
"""

from __future__ import annotations

from capellambse import model as m

from . import capellacommon, capellacore, cs, fa, information, interaction


@m.xtype_handler(None)
class OperationalActivity(fa.AbstractFunction):
    """An operational activity."""

    _xmltag = "ownedOperationalActivities"

    exchanges = m.DirectProxyAccessor(
        fa.FunctionalExchange, aslist=m.ElementList
    )

    inputs = m.ReferenceSearchingAccessor(
        fa.FunctionalExchange, "target", aslist=m.ElementList
    )
    outputs = m.ReferenceSearchingAccessor(
        fa.FunctionalExchange, "source", aslist=m.ElementList
    )

    owner: m.Accessor[Entity]

    @property
    def related_exchanges(self) -> m.ElementList[fa.FunctionalExchange]:
        seen: set[str] = set()
        exchanges = []
        for fex in self.inputs + self.outputs:
            if fex.uuid not in seen:
                exchanges.append(fex._element)
                seen.add(fex.uuid)
        return self.inputs._newlist(exchanges)


@m.xtype_handler(None)
class OperationalProcess(fa.FunctionalChain):
    """An operational process."""


@m.xtype_handler(None)
class EntityOperationalCapabilityInvolvement(interaction.AbstractInvolvement):
    """An EntityOperationalCapabilityInvolvement."""


@m.xtype_handler(None)
class OperationalCapability(m.ModelElement):
    """A capability in the OperationalAnalysis layer."""

    _xmltag = "ownedOperationalCapabilities"

    extends = m.DirectProxyAccessor(
        interaction.AbstractCapabilityExtend, aslist=m.ElementList
    )
    extended_by = m.ReferenceSearchingAccessor(
        interaction.AbstractCapabilityExtend, "target", aslist=m.ElementList
    )
    includes = m.DirectProxyAccessor(
        interaction.AbstractCapabilityInclude, aslist=m.ElementList
    )
    included_by = m.ReferenceSearchingAccessor(
        interaction.AbstractCapabilityInclude, "target", aslist=m.ElementList
    )
    generalizes = m.DirectProxyAccessor(
        interaction.AbstractCapabilityGeneralization, aslist=m.ElementList
    )
    generalized_by = m.DirectProxyAccessor(
        interaction.AbstractCapabilityGeneralization,
        "target",
        aslist=m.ElementList,
    )
    involved_activities = m.LinkAccessor[OperationalActivity](
        "ownedAbstractFunctionAbstractCapabilityInvolvements",
        interaction.AbstractFunctionAbstractCapabilityInvolvement,
        aslist=m.ElementList,
        attr="involved",
    )
    involved_entities = m.LinkAccessor[m.ModelElement](
        "ownedEntityOperationalCapabilityInvolvements",
        EntityOperationalCapabilityInvolvement,
        aslist=m.MixedElementList,
        attr="involved",
    )
    entity_involvements = m.DirectProxyAccessor(
        EntityOperationalCapabilityInvolvement, aslist=m.ElementList
    )
    involved_processes = m.LinkAccessor[OperationalProcess](
        "ownedFunctionalChainAbstractCapabilityInvolvements",
        interaction.FunctionalChainAbstractCapabilityInvolvement,
        aslist=m.ElementList,
        attr="involved",
    )
    owned_processes = m.DirectProxyAccessor(
        OperationalProcess, aslist=m.ElementList
    )

    postcondition = m.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = m.AttrProxyAccessor(capellacore.Constraint, "preCondition")
    scenarios = m.DirectProxyAccessor(
        interaction.Scenario, aslist=m.ElementList
    )
    states = m.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=m.ElementList
    )

    packages: m.Accessor


@m.xtype_handler(None)
class OperationalCapabilityPkg(m.ModelElement):
    """A package that holds operational capabilities."""

    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = m.DirectProxyAccessor(
        OperationalCapability, aslist=m.ElementList
    )

    packages: m.Accessor


class AbstractEntity(cs.Component):
    """Common code for Entities."""

    activities = m.LinkAccessor[OperationalActivity](
        "ownedFunctionalAllocation",
        fa.ComponentFunctionalAllocation,
        aslist=m.ElementList,
        attr="targetElement",
        backattr="sourceElement",
    )
    capabilities = m.ReferenceSearchingAccessor(
        OperationalCapability, "involved_entities", aslist=m.ElementList
    )


@m.xtype_handler(None)
class Entity(AbstractEntity):
    """An Entity in the OperationalAnalysis layer."""

    _xmltag = "ownedEntities"

    entities: m.Accessor

    @property
    def inputs(self) -> m.ElementList[CommunicationMean]:
        return self._model.search(CommunicationMean).by_target(self)

    @property
    def outputs(self) -> m.ElementList[CommunicationMean]:
        return self._model.search(CommunicationMean).by_source(self)


@m.xtype_handler(None)
class OperationalActivityPkg(m.ModelElement):
    """A package that holds operational entities."""

    _xmltag = "ownedFunctionPkg"

    activities = m.DirectProxyAccessor(
        OperationalActivity, aslist=m.ElementList
    )

    packages: m.Accessor


@m.xtype_handler(None)
class CommunicationMean(fa.AbstractExchange):
    """An operational entity exchange."""

    _xmltag = "ownedComponentExchanges"

    allocated_interactions = m.LinkAccessor[fa.FunctionalExchange](
        None,  # FIXME fill in tag
        fa.ComponentExchangeFunctionalExchangeAllocation,
        aslist=m.ElementList,
        attr="targetElement",
    )
    allocated_exchange_items = m.AttrProxyAccessor(
        information.ExchangeItem,
        "convoyedInformations",
        aslist=m.ElementList,
    )

    exchange_items = fa.ComponentExchange.exchange_items


@m.xtype_handler(None)
class EntityPkg(m.ModelElement):
    """A package that holds operational entities."""

    _xmltag = "ownedEntityPkg"

    entities = m.DirectProxyAccessor(Entity, aslist=m.ElementList)
    state_machines = m.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=m.ElementList
    )

    packages: m.Accessor
    exchanges = m.DirectProxyAccessor(CommunicationMean, aslist=m.ElementList)


@m.xtype_handler(None)
class OperationalAnalysis(cs.ComponentArchitecture):
    """Provides access to the OperationalAnalysis layer of the model."""

    root_entity = m.DirectProxyAccessor(Entity, rootelem=EntityPkg)
    root_activity = m.DirectProxyAccessor(
        OperationalActivity, rootelem=OperationalActivityPkg
    )

    activity_package = m.DirectProxyAccessor(OperationalActivityPkg)
    capability_package = m.DirectProxyAccessor(OperationalCapabilityPkg)
    entity_package = m.DirectProxyAccessor(EntityPkg)

    all_activities = m.DeepProxyAccessor(
        OperationalActivity,
        aslist=m.ElementList,
    )
    all_processes = m.DeepProxyAccessor(
        OperationalProcess,
        aslist=m.ElementList,
    )
    all_capabilities = m.DeepProxyAccessor(
        OperationalCapability,
        aslist=m.ElementList,
    )
    all_actors = property(
        lambda self: self._model.search(Entity).by_is_actor(True)
    )
    all_entities = m.DeepProxyAccessor(
        Entity,
        aslist=m.ElementList,
    )

    all_activity_exchanges = m.DeepProxyAccessor(
        fa.FunctionalExchange,
        aslist=m.ElementList,
        rootelem=[OperationalActivityPkg, OperationalActivity],
    )
    all_entity_exchanges = m.DeepProxyAccessor(
        CommunicationMean,
        aslist=m.ElementList,
    )
    all_operational_processes = property(
        lambda self: self._model.search(OperationalProcess, below=self)
    )

    diagrams = m.DiagramAccessor(
        "Operational Analysis", cacheattr="_MelodyModel__diagram_cache"
    )


m.set_accessor(
    OperationalActivity,
    "packages",
    m.DirectProxyAccessor(OperationalActivityPkg, aslist=m.ElementList),
)
m.set_accessor(
    OperationalActivity,
    "owner",
    m.ReferenceSearchingAccessor(Entity, "activities"),
)
m.set_accessor(
    Entity,
    "exchanges",
    m.DirectProxyAccessor(CommunicationMean, aslist=m.ElementList),
)
m.set_accessor(
    Entity,
    "related_exchanges",
    m.ReferenceSearchingAccessor(
        CommunicationMean, "source", "target", aslist=m.ElementList
    ),
)
m.set_self_references(
    (OperationalActivity, "activities"),
    (OperationalActivityPkg, "packages"),
    (OperationalCapabilityPkg, "packages"),
    (Entity, "entities"),
    (EntityPkg, "packages"),
)



#!/usr/bin/env -S python -Xdev -Xfrozen_modules=off
# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""A simple Python REPL with a Capella model.

You can use this script to quickly launch an interactive Python
interpreter and load a model.

If the ``readline`` module is available, the interpreter spawned from
this script uses a separate readline history. It is located in
``$XDG_STATE_HOME/capellambse`` on proper operating systems and in the
``capellambse`` cache directory on others.

Normally this script is run with something like:

.. code-block:: bash

   python -Xdev -m capellambse.repl test-5.0

However, as that can become quite unwieldy, it is also possible to run
it directly from the source tree with:

.. code-block:: bash

    ./capellambse/repl.py test-5.0

On Unix-like operating systems, this will automatically enable ``-Xdev``
on the Python interpreter.

In order to add custom models for the :ref:`--model <repl.py-model>`
argument, add a JSON file to your user ``known_models`` directory. This
file defines the instantiation parameters for the
:class:`capellambse.model.MelodyModel`:

.. literalinclude:: ../../../capellambse/known_models/test-lib.json
   :language: json
   :lineno-start: 1
   :linenos:

Use the following command to view the path of your ``known_models``
directory, and to see which models are currently available:

.. code-block:: bash

   python -m capellambse.cli_helpers

In addition to the standard Python builtins, he environment in the REPL
also provides the following convenience imports:

- The Python packages ``capellambse``, ``inspect``, ``logging``, ``os``,
  ``pathlib``
- ``importlib`` as ``im``, with its submodules ``importlib.metadata`` as
  ``imm`` and ``importlib.resources`` as ``imr``
- The ``etree`` module from ``lxml``
- The ``pprint`` function from the ``pprint`` module

A few helpful functions specific to working interactively with models
are also available. Use the ``help`` builtin inside the REPL to get more
information and usage examples.

- ``fzf``: Wrapper around the ``fzf`` binary (must be in the PATH) which
  makes it easy to select a model element interactively from a list
- ``logtee``: Context manager that redirects log messages to a file
- ``showxml``: Print the XML representation of a model object
- ``suppress``: Context manager that suppresses exceptions of given type
"""

from __future__ import annotations

import argparse
import code
import collections.abc as cabc
import contextlib
import importlib
import json
import logging
import operator
import os
import os.path
import pathlib
import shutil
import subprocess
import sys
import textwrap
import typing as t

from lxml import etree

import capellambse
from capellambse import cli_helpers
from capellambse.loader import exs

try:
    import readline

    with contextlib.suppress(ImportError):
        import rlcompleter  # noqa: F401

    class _ReadlineHistory:
        def __init__(self, histfile):
            self.histfile = histfile

        def __enter__(self) -> None:
            with contextlib.suppress(FileNotFoundError):
                readline.read_history_file(self.histfile)
            readline.parse_and_bind("tab: complete")

        def __exit__(self, *_) -> None:
            try:
                self.histfile.parent.mkdir(parents=True, exist_ok=True)
                self.histfile.open("wb").close()
            except OSError:
                pass
            readline.append_history_file(100_000, self.histfile)

except ImportError:

    class _ReadlineHistory:  # type: ignore[no-redef]
        def __init__(self, *__, **_):
            pass

        def __enter__(self):
            pass

        def __exit__(self, *_):
            pass


basedir = pathlib.Path(__file__).parent.resolve()
logger = logging.getLogger("capellambse.repl")


def _parse_args() -> dict[str, t.Any] | None:
    known_models = {
        i.name[: -len(".json")]: i
        for i in capellambse.enumerate_known_models()
        if i.name.endswith(".json")
    }
    parser = argparse.ArgumentParser("capellambse/repl.py")
    parser.add_argument(
        "model",
        default=None,
        nargs="?",
        help=(
            "A model name from known_models, an AIRD file, or the path to (or"
            " contents of) a JSON file describing the model. The following"
            " models are known: " + ", ".join(f"``{i}``" for i in known_models)
        ),
    )
    parser.add_argument(
        "--disable-diagram-cache",
        action="store_true",
        help="Disable the diagram cache, if one was defined for the model",
    )
    parser.add_argument(
        "--dump",
        action="store_true",
        help="Dump model info as JSON to stdout and exit",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"capellambse {capellambse.__version__}",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--hold",
        action="store_true",
        help="Inhibit automatic model updates (update_cache=False)",
    )
    group.add_argument(
        "--wipe",
        action="store_true",
        help="Wipe the cache (disable_cache=True)",
    )

    pargs = parser.parse_args(sys.argv[1:])
    if pargs.model is None:
        if pargs.dump:
            raise SystemExit("No model specified, cannot dump")
        return None

    modelinfo = cli_helpers.loadinfo(pargs.model)

    if pargs.hold:
        modelinfo["update_cache"] = False
    if pargs.wipe:
        modelinfo["disable_cache"] = True

    if pargs.disable_diagram_cache and "diagram_cache" in modelinfo:
        del modelinfo["diagram_cache"]

    if pargs.dump:
        print(json.dumps(modelinfo, indent=2))
        raise SystemExit(0)
    return modelinfo


@contextlib.contextmanager
def logtee(
    filename: os.PathLike[str] | str,
    /,
    *,
    append: bool = False,
    catch: bool = False,
    tee: bool = True,
    level: int | str | None = None,
    logger: str = "",
) -> cabc.Generator[None, None, None]:
    """Temporarily write log messages to a file.

    This function adds an additional log handler, meaning that the
    current logging targets (usually stderr) will also still be served.

    Parameters
    ----------
    filename
        The file path to write log messages to.
    append
        Append to the file instead of overwriting it.
    catch
        If ``True``, catches and suppresses exceptions that reach the
        context manager. Only ``Exception`` and subtypes will be
        suppressed. Regardless of this parameter, exceptions (of any
        type) will always be logged (subject to the log level).
    level
        Change the log level during redirection.
    logger
        Attach to this sub-logger instead of the root logger.
    tee
        Keep the original log destination(s) intact as well. If
        ``False``, also inhibits propagation of log messages to parent
        loggers.
    """
    loggerobj = logging.getLogger(logger)
    handler = logging.FileHandler(filename, mode="wa"[bool(append)])

    orig_handlers = loggerobj.handlers
    orig_propagate = loggerobj.propagate
    orig_level = loggerobj.level

    loggerobj.handlers = orig_handlers * bool(tee) + [handler]
    loggerobj.propagate = bool(tee)
    if level is not None:
        loggerobj.setLevel(level)

    try:
        yield
    except BaseException:
        loggerobj.handlers = [handler]
        loggerobj.propagate = False
        loggerobj.exception("Exception reached top of context")
        if not catch:
            raise
    finally:
        loggerobj.handlers = orig_handlers
        loggerobj.propagate = orig_propagate
        loggerobj.level = orig_level
        handler.close()


@contextlib.contextmanager
def suppress(
    *exc_types: type[BaseException], log: bool = True
) -> cabc.Iterator[None]:
    """Prevent the specified types of exceptions from propagating.

    Parameters
    ----------
    exc_types
        The exception types to suppress. Subclasses of these exceptions
        will also be suppressed.
    log
        Print a short warning about the exception to stderr.

    Examples
    --------
    >>> with suppress(ValueError):
    ...     print("Hello")
    ...     raise ValueError("oops")
    ...     print("World")
    ...
    Hello
    """
    try:
        yield
    except exc_types:
        if log:
            logging.info("Exception suppressed", exc_info=True)


def showxml(obj: capellambse.ModelObject | etree._Element) -> None:
    """Show the XML representation of a model object.

    Examples
    --------
    >>> my_obj = model.search("LogicalComponent").by_name("My Component")
    >>> showxml(my_obj)
    <ownedLogicalComponents name="My Component">
      ...
    </ownedLogicalComponents>
    """
    if isinstance(obj, etree._Element):
        elm = obj
    else:
        elm = obj._element
    print(exs.to_string(elm), end="")


def fzf(
    elements: cabc.Iterable[capellambse.ModelObject],
    attr: str = "name",
) -> capellambse.ModelObject | None:
    """Interactively select an element using fzf.

    This function requires the ``fzf`` binary to be installed and in the
    ``$PATH``.

    Examples
    --------
    >>> # Select a LogicalComponent by name
    >>> obj = fzf(model.search("LogicalComponent"))

    >>> # Select a ComponentExchange by the name of its target component
    >>> obj = fzf(model.search("ComponentExchange"), "target.parent.name")
    """

    def repr(obj):
        return getattr(obj, "_short_repr_", obj.__repr__)()

    binary = shutil.which("fzf")
    if not binary:
        raise RuntimeError("fzf is not installed")
    elements = list(elements)

    getter = operator.attrgetter(attr)

    entries = [(str(getter(i)).replace("\0", ""), repr(i)) for i in elements]
    maxlen = max(len(i[0]) for i in entries)
    maxlen = min(maxlen, 40)
    fzf_input = "\0".join(
        f"{i} \x1b[97m{s:{maxlen}}  \x1b[36m{e}"
        for i, (s, e) in enumerate(entries)
    )

    try:
        proc = subprocess.run(
            [binary, "--ansi", "--read0", "--with-nth=2.."],
            check=True,
            input=fzf_input,
            text=True,
            stdout=subprocess.PIPE,
        )
    except (Exception, KeyboardInterrupt):
        return None
    else:
        selected = elements[int(proc.stdout.strip().split(" ", 1)[0])]
        print(repr(selected))
        return selected


def main() -> None:
    """Launch a simple Python REPL with a Capella model."""
    os.chdir(pathlib.Path(capellambse.__file__).parents[1])

    interactive_locals = {
        "__doc__": None,
        "__name__": "__console__",
        "etree": etree,
        "fzf": fzf,
        "im": importlib,
        "imm": importlib.import_module("importlib.metadata"),
        "imr": importlib.import_module("importlib.resources"),
        "logtee": logtee,
        "pprint": importlib.import_module("pprint").pprint,
        "showxml": showxml,
        "suppress": suppress,
    }

    for m in ("capellambse", "inspect", "logging", "os", "pathlib"):
        interactive_locals[m] = importlib.import_module(m)

    modelinfo = _parse_args()
    if modelinfo is not None:
        logger.debug("Loading model: %r", modelinfo)
        model = capellambse.MelodyModel(**modelinfo)
        interactive_locals["model"] = model
        banner = textwrap.dedent(
            f"""\

            {' Model exploration ':=^80}
            `model` is {model.info.title!r}
            from {modelinfo['path']}
            """
        )
    else:
        banner = textwrap.dedent(
            f"""\

            {' Model exploration ':=^80}
            Load a model with `model = capellambse.loadcli("model-name")`
            or `model = capellambse.MelodyModel("uri", arg=...)`.
            """
        )

    banner += textwrap.dedent(
        """\

        Convenience imports:
        - `capellambse`, `inspect`, `logging`, `os`, `pathlib`
        - `im` = importlib (`imm` = .metadata, `imr` = .resources)
        - `etree` = lxml.etree, `pprint` = pprint.pprint

        Helpful functions and context managers (use `help(name)`):
        - `fzf`: Select a model element interactively from a list
        - `logtee`: CM that redirects log messages to a file
        - `showxml`: Print the XML representation of a model object
        - `suppress`: CM that suppresses exceptions of given type
        """
    )

    history_file = capellambse.dirs.user_state_path / "model_exploration.hist"
    with _ReadlineHistory(history_file), suppress(BrokenPipeError):
        code.interact(banner=banner, local=interactive_locals, exitmsg="")


if __name__ == "__main__":
    main()
